import re, json

with open('data/compiled_regexs.json') as f:
    regexs = json.load(f)['py']

def myassert(b, should_be, i, failing, e=None):
    if bool(b) != should_be:
        print(f'''
----------------------- TEST FAILED -----------------------
language       = `py`
pattern        = `{i['ezregex']}`
compiled regex = `{i['regex']}`
pattern should {'' if should_be else 'NOT '}match `{failing.replace('\n', '\\n')}`
''')
        if e:
            raise e
        exit(1)

for i in regexs:
    for m in i['should']:
        try:
            myassert(re.search(i['regex'], m), True, i, m)
        except Exception as e:
            myassert(False, True, i, m, e)
    for m in i['shouldnt']:
        try:
            myassert(re.search(i['regex'], m), False, i, m)
        except Exception as e:
            myassert(True, False, i, m, e)

with open('data/compiled_replacements.json') as f:
    replacements = json.load(f)['py']

def repl_assert(i, actual):
    if actual != i['after']:
        print(f'''
----------------------- TEST FAILED -----------------------
language       = `py`
pattern        = `{i['ezregex']}`
compiled regex = `{i['regex']}`
replacement    = `{i['ezrepl']}`
compiled repl  = `{i['repl']}`
base           = `{i['base']}`
after          = `{i['after']}`

Replacing
    `{i["ezregex"]}`
with
    `{i["repl"]}`
in
    `{i["base"]}`
yielded
    `{actual}`
not
    `{i["after"]}`
''')
        exit(1)

for i in replacements:
    try:
        repl_assert(i, re.sub(i['regex'], i['repl'], i['base']))
    except Exception as e:
        print(f'Error replacing `{i["ezregex"]}` with `{i["repl"]}` in `{i["base"]}`: {str(e)}')
        exit(1)

print('pass')


#!/usr/bin/env python3
""" A readable and intuitive way to generate Regular Expressions """
__version__ = '2.3.0'

# Import this as a submodule
from . import generate
from .api import api
from .EZRegex import EZRegex
from .generate import generate_regex
from .invert import invert
# Python is the default
from .python import *
from . import R
from . import javascript
from . import PCRE2
# from ._docs import operator_docs as _operator_docs
# from ._docs import docs as _docs



import threading
import time

from rich import print as rprint
from rich.table import Table
from rich.text import Text
from rich.progress import track

from ezregex import invert
import re
import json

import argparse

parser = argparse.ArgumentParser()
parser.add_argument('--strictness', type=int, default=20)
parser.add_argument('--tries', type=int, default=5)
parser.add_argument('--timeout', type=int, default=5)
parser.add_argument('--backend', type=str, default=...)
parser.add_argument('--passed', action='store_true')
args = parser.parse_args()

# TODO: make these cli params
strictness=args.strictness
includePassed=args.passed
invertBackend=args.backend
# -1 means return it even if it's bad
invert_tries=args.tries
timeout_limit = args.timeout * args.strictness

# The structure is ["regex pattern", "replacement regex", "base string", "what the base string should look like after substitution"]
with open('data/compiled_replacements.json') as f:
    replacements = json.load(f)['py']

with open('data/compiled_regexs.json') as f:
    regexs = json.load(f)['py']

# with open('deleteme.txt', 'w') as f:
#     rprint(regexs + replacements, file=f)
table = Table(title="invert tests", expand=False)
# table.add_column("Line", justify="center", style="grey37")#, max_width=2)
table.add_column("Regex", justify="right", style="green")#, max_width=40)
table.add_column("Inverse", justify="left", style="dim green")
table.add_column("Status", justify="center")

res = [i['regex'] for i in regexs + replacements]
failures = []
for r in track(res, total=len(res)):
    def do_test(r):
        global failures
        # For SOME REASON this one AND ONLY THIS ONE doesn't work INSIDE the docker container, but
        # works OUTSIDE of it. I have no idea why. They've each been tested manually.
        if r in (
            r'(<)?(\w+@\w+(?:\.\w+)+)(?(1)>|$)',
            r'(?:(<))?(\w+@\w+(?:\.\w+)+)(?(1)>|\Z)',
        ):
            return
        try:
            for _ in range(strictness):
                inv = invert(r, backend=invertBackend, tries=invert_tries)
                if re.search(r, inv) is None:
                    failures.append((r, inv, 'failed'))
                elif includePassed:
                    failures.append((r, inv, 'passed'))
        except (Exception, AssertionError) as err:
            failures.append((r, str(err), 'error'))
            raise err

    # print(r, end='\n')
    thread = threading.Thread(target=do_test, args=(r,))
    thread.start()
    thread.join(timeout=timeout_limit)
    # assert not thread.is_alive(), f"Invert took longer than {timeout_limit} seconds: `{r}`"
    if thread.is_alive():
        failures.append((r, '', 'timeout'))
    # Murder the thread
    thread._stop()

# skipped = set()
for r, inv, status in failures:
    # if r in skipped:
        # continue
    # If there's `strictness` number of the same failure, just say they all failed
    # This doesn't work. Don't know why.
    # if len([i for i in failures if i[0] == r]) == strictness:
    #     skipped.add(r)
    table.add_row(
        Text(r),
        Text('`' + inv + '`', style='dim green'),
        Text(status, style='red' if status != 'passed' else 'blue')
    )

if len(failures):
    with open('summary.txt', 'w') as f:
        rprint(table, file=f)
    rprint(table)

print(f'{len(failures)}/{len(res)} ({len(failures)/len(res)*100:.2f}%) failed')

assert not len(failures)

print('pass')


import re
import sys
from functools import partial
from typing import Any, Callable, Iterator, Self

from mypy_extensions import DefaultNamedArg, VarArg

from .base.interface import InputType

type EZRegexDefinition = str|Callable[[VarArg, DefaultNamedArg[str, "cur"]], str]|list[partial[str]]

class EZRegex:
    """ Represent parts of the Regex syntax. Should not be instantiated by the user directly."""
    _sanatize: bool
    _flags: str
    # _funcList: list[Callable[[VarArg, DefaultNamedArg[str, "cur"]], str]]
    _funcList: list
    replacement: bool

    def __init__(self, definition:EZRegexDefinition, *, sanatize:bool=True, replacement:bool=False, flags:str='') -> None:
        """ The workhorse of the EZRegex library. This represents a regex pattern that can be combined
            with other EZRegexs and strings. Ideally, this should only be called internally, but it should
            still work from the user's end
        """

    # Private functions
    def _flag_func(self, final:str) -> str: ...
    def _escape(self, pattern:str) -> str:
        """ This function was modified from the one in /usr/lib64/python3.12/re/__init__.py line 255 """
    def _sanitizeInput(self, i:InputType, addFlags:bool=False) -> str:
        """ Instead of rasising an error if passed a strange datatype, it now trys to cast it to a string """
    def _compile(self, addFlags=True) -> str: ...
    def _copy(self, definition:EZRegexDefinition=..., sanatize:bool=..., replacement:bool=..., flags:str=...): ...

    # Regular functions
    def compile(self, addFlags=True) -> re.Pattern: ...
    def str(self) -> str: ...
    def debug(self) -> EZRegex: ...
    def copy(self, addFlags=True) -> None: ...
    def test(self, testString:str=None, show=True, context=True) -> bool:
        """ Tests the current regex expression to see if it's in @param testString.
            Returns the match objects (None if there was no match)
        """
    def invert(self, amt=1, **kwargs) -> str: ...
    def inverse(self, amt=1, **kwargs) -> str:
        """ "Inverts" the current Regex expression to give an example of a string it would match.
            Useful for debugging purposes. """

    # Elemental functions
    def group(self, name:str=None) -> EZRegex: ...
    def named(self, name:str) -> EZRegex: ...
    @property
    def unnamed(self) -> EZRegex: ...
    def if_not_preceded_by(self, input:InputType) -> EZRegex: ...
    def if_preceded_by(self, input:InputType) -> EZRegex: ...
    def if_not_proceded_by(self, input:InputType) -> EZRegex: ...
    def if_proceded_by(self, input:InputType) -> EZRegex: ...
    def if_not_followed_by(self, input:InputType) -> EZRegex: ...
    def if_followed_by(self, input:InputType) -> EZRegex: ...
    def if_enclosed_with(self, open:str, closed:str|None=None) -> EZRegex: ...
    @property
    def optional(self) -> EZRegex: ...
    @property
    def repeat(self) -> EZRegex: ...
    @property
    def exactly(self) -> EZRegex: ...
    def at_least(self, min:int) -> EZRegex: ...
    def more_than(self, min:int) -> EZRegex: ...
    def amt(self, amt:int) -> EZRegex: ...
    def at_most(self, max:int) -> EZRegex: ...
    def between(self, min:int, max:int, greedy:bool=True, possessive:bool=False) -> EZRegex: ...
    def at_least_one(self, greedy:bool=True, possessive:bool=False) -> EZRegex: ...
    def at_least_none(self, greedy:bool=True, possessive:bool=False) -> EZRegex: ...
    def or_(self, input:InputType) -> EZRegex: ...
    @property
    def ASCII(self) -> EZRegex: ...
    @property
    def IGNORECASE(self) -> EZRegex: ...
    @property
    def DOTALL(self) -> EZRegex: ...
    @property
    def LOCALE(self) -> EZRegex: ...
    @property
    def MULTILINE(self) -> EZRegex: ...
    @property
    def UNICODE(self) -> EZRegex: ...

    # Named operator functions
    def append(self, input:InputType) -> EZRegex: ...
    def prepend(self, input:InputType) -> EZRegex: ...

    # Flag functions
    @property
    def flags(self) -> str: ...
    def set_flags(self, to:str) -> EZRegex: ...
    def add_flag(self, flag:str) -> EZRegex: ...
    def remove_flag(self, flag:str) -> EZRegex: ...

    # Magic Functions
    def __call__(self, *args, **kwargs) -> EZRegex | str:
        """ This should be called by the user to specify the specific parameters of this instance i.e. anyof('a', 'b') """
    def __str__(self, addFlags:bool=True) -> str: ...
    def __repr__(self) -> str: ...
    def __eq__(self, thing:InputType) -> bool: ...
    def __mul__(self, amt:int) -> EZRegex: ...
    def __rmul__(self, amt:int) -> EZRegex: ...
    def __imul__(self, amt:int) -> EZRegex: ...
    def __add__(self, thing:InputType) -> EZRegex: ...
    def __radd__(self, thing:InputType) -> EZRegex: ...
    def __iadd__(self, thing:InputType) -> EZRegex: ...
    def __and__(self, thing:InputType) -> EZRegex: ...
    def __rand__(self, thing:InputType) -> EZRegex: ...
    # The shift operators just shadow the add operators
    def __lshift__(self, thing:InputType) -> EZRegex: ...
    def __rlshift__(self, thing:InputType) -> EZRegex: ...
    def __ilshift__(self, thing:InputType) -> EZRegex: ...
    # I don't think right and left shifts should be any different, right?
    def __rshift__(self, thing:InputType) -> EZRegex: ...
    def __rrshift__(self, thing:InputType) -> EZRegex: ...
    def __irshift__(self, thing:InputType) -> EZRegex: ...
    def __invert__(self) -> str: ...
    def __pos__(self) -> EZRegex:
        """ TODO: Add documentation here """
    def __ror__(self, thing:InputType) -> EZRegex:
        """ TODO: Add documentation here """
    def __or__(self, thing:InputType) -> EZRegex: ...
    def __xor__(self, thing:InputType) -> EZRegex: ...
    def __rxor__(self, thing:InputType) -> EZRegex: ...
    def __mod__(self, other:str) -> re.Match|None:
        """ I would prefer __rmod__(), but it doesn't work on strings, since __mod__() is already specified for string formmating. """
    def __hash__(self) -> int: ...
    def __contains__(self, thing:str) -> bool: ...
    def __getitem__(self, args) -> EZRegex: ...
    def __reversed__(self) -> str: ...
    def __rich__(self) -> str: ...
    def __pretty__(self) -> str: ...
    # To make EZRegex instances immutable
    def __setattr__(self, name:str, value, ignore=False): ...
    def __delattr__(self, *args): ...



#!/usr/bin/env python3
from ..EZRegex import EZRegex

# This file gets parsed to generate the documentation, and to provide descriptions and things to ezregex.org
# You should be able to treat it as any .pyi file, with the addition that strings of the form
# "Group: <name>\n<optional description>"
# designate the group for that section, and strings below variables act as the descriptions for those variables.

InputType = str | EZRegex | int | float


string_starts_with: EZRegex
string_ends_with: EZRegex
line_starts_with: EZRegex
line_ends_with: EZRegex
word_boundary: EZRegex
"Matches the boundary of a word, i.e. the empty space between a word character and not a word character, or the end of a string"
not_word_boundary: EZRegex
"The opposite of `wordBoundary`"


tab: EZRegex
space: EZRegex
space_or_tab: EZRegex
new_line: EZRegex
carriage_return: EZRegex
quote: EZRegex
"Matches ', \", and `"
vertical_tab: EZRegex
form_feed: EZRegex
comma: EZRegex
period: EZRegex
underscore: EZRegex


not_whitespace: EZRegex
not_digit: EZRegex
not_word: EZRegex


whitespace: EZRegex
whitechunk: EZRegex
"A \"chunk\" of whitespace. Just any amount of whitespace together"
digit: EZRegex
number: EZRegex
"Matches multiple digits next to each other. Does not match negatives or decimals"
word: EZRegex
word_char: EZRegex
"Matches just a single \"word character\", defined as any letter, number, or _"
anything: EZRegex
"Matches any single character, except a newline. To also match a newline, use literally_anything"
chunk: EZRegex
'A "chunk": Any clump of characters up until the next newline'
uppercase: EZRegex
lowercase: EZRegex
letter: EZRegex
"Matches just a letter -- not numbers or _ like word_char"
hex_digit: EZRegex
oct_digit: EZRegex
punctuation: EZRegex
controller: EZRegex
"Matches a metadata ASCII characters"
printable: EZRegex
"Matches printable ASCII characters"
printable_and_space: EZRegex
alpha_num: EZRegex
unicode: EZRegex
"Matches a unicode character by name"

def any_between(char:str, and_char:str) -> EZRegex:
    "Match any char between `char` and `and_char`, using the ASCII table for reference"
    ...


def match_max(input:InputType) -> EZRegex:
    """ Match as many of `input` in the string as you can. This is equivelent to using the unary + operator.
    If `input` is not provided, it works on the previous regex pattern. That's not recommended for
    clarity's sake though
    """
    ...

def match_num(num: int, input: InputType) -> EZRegex:
    "Match `num` amount of `input` in the string"
    ...

def more_than(min: int, input: InputType) -> EZRegex:
    "Match more than `min` sequences of `input` in the string"
    ...

def at_least(min:int, input:InputType) -> EZRegex:
    "Match at least `min` sequences of `input` in the string"
    ...

def at_most(max:int, input:InputType) -> EZRegex:
    "Match at most `max` instances of `input` in the string"
    ...

def between(min:int, max:int, input:InputType, greedy:bool=True, possessive:bool=False) -> EZRegex:
    """ Match between `min` and `max` sequences of `input` in the string. This also accepts `greedy` and `possessive` parameters
        Max can be an empty string to indicate no maximum
        `greedy` means it will try to match as many repititions as possible
        non-greedy will try to match as few repititions as possible
        `possessive` means it won't backtrack to try to find any repitions
        see https://docs.python.org/3/library/re.html for more help
    """
    ...

def at_least_one(input:InputType, greedy:bool=True, possessive:bool=False) -> EZRegex:
    """ Match at least one of `input` in the string. This also accepts `greedy` and `possessive` parameters
        `greedy` means it will try to match as many repititions as possible
        non-greedy will try to match as few repititions as possible
        `possessive` means it won't backtrack to try to find any repitions
        see https://docs.python.org/3/library/re.html for more help
    """
    ...

def at_least_none(input:InputType, greedy:bool=True, possessive:bool=False) -> EZRegex:
    """ Match 0 or more sequences of `input`. This also accepts `greedy` and `possessive` parameters
        `greedy` means it will try to match as many repititions as possible
        non-greedy will try to match as few repititions as possible
        `possessive` means it won't backtrack to try to find any repitions
        see https://docs.python.org/3/library/re.html for more help
    """
    ...


def optional(input:InputType, greedy:bool=True, possessive:bool=False) -> EZRegex:
    """ Match `input` if it's there. This also accepts `greedy` and `possessive` parameters
        `greedy` means it will try to match as many repititions as possible
        non-greedy will try to match as few repititions as possible
        `possessive` means it won't backtrack to try to find any repitions
        see https://docs.python.org/3/library/re.html for more help
    """
    ...

def either(input:InputType, or_input:InputType) -> EZRegex: ...

def any_of(*inputs:str, chars:bool|None=None, split:bool|None=None) -> EZRegex:
    """ Match any of the given `inputs`. Note that `inputs` can be multiple parameters,
        or a single string. Can also accept parameters chars and split. If char is set
        to True, then `inputs` must only be a single string, it interprets `inputs`
        as characters, and splits it up to find any of the chars in the string. If
        split is set to true, it forces the ?(...) regex syntax instead of the [...]
        syntax. It should act the same way, but your output regex will look different.
        By default, it just optimizes it for you.
    """
    ...

def any_char_except(*inputs:str) -> EZRegex:
    "This matches any char that is NOT in `inputs`. `inputs` can be multiple parameters, or a single string of chars to split."
    ...

def any_except(input:InputType, type:InputType='.*') -> EZRegex:
    """ Matches anything other than `input`, which must be a single string or EZRegex chain, **not** a list. Also
    optionally accepts the `type` parameter, which works like this: \"Match any `type` other than `input`\". For example,
    \"match any word which is not foo\". Do note that this function is new, and I'm still working out the kinks.
    """
    ...

def each(*inputs:InputType) -> EZRegex:
    "Matches if the next part of the string can match all of the given inputs. Like the + operator, but out of order."
    ...



def if_proceded_by(input:InputType) -> EZRegex:
    """ Matches the pattern if it has `input` coming after it. Can only be used once in a given pattern,
        as it only applies to the end
    """
    ...

def if_not_proceded_by(input:InputType) -> EZRegex:
    """ Matches the pattern if it does **not** have `input` coming after it. Can only be used once in
        a given pattern, as it only applies to the end
    """
    ...

def if_preceded_by(input:InputType) -> EZRegex:
    """ Matches the pattern if it has `input` coming before it. Can only be used once in a given pattern,
        as it only applies to the beginning
    """
    ...

def if_not_preceded_by(input:InputType) -> EZRegex:
    """ Matches the pattern if it does **not** have `input` coming before it. Can only be used once
        in a given pattern, as it only applies to the beginning
    """
    ...

def if_enclosed_with(open:str, stuff:InputType, close:str|None=None) -> EZRegex:
    """ Matches if the string has `open`, then `stuff`, then `close`, but only \"matches\"
        stuff. Just a convenience combination of ifProceededBy and ifPreceededBy.
    """
    ...


def group(input:InputType, name:str|None=None) -> EZRegex:
    "Causes `input` to be captured as an unnamed group. Only useful when replacing regexs"
    ...

def passive_group(input:InputType) -> EZRegex:
    "As all regexs in EZRegex capture passively, this is entirely useless. But if you really want to, here it is"
    ...

def earlier_group(num_or_name:int|str) -> EZRegex:
    """ Matches whatever the group referenced by `num_or_name` matched earlier. Must be *after* a
    group which would match `num_or_name`
    """
    ...

def if_exists(num_or_name:int|str, does:InputType, doesnt:InputType|None=None) -> EZRegex:
    """ Matches `does` if the group `num_or_name` exists, otherwise it matches `doesnt` """
    ...



def rgroup(num_or_name:str|int) -> EZRegex:
    """ Puts in its place the group specified, either by group number (for unnamed
        groups) or group name (for named groups). Named groups are also counted by
        number, I'm pretty sure. Groups are numbered starting from 1
    """
    ...

replace_entire: EZRegex
"Puts in its place the entire match"

def replace(string:str, rtn_str:bool=True) -> str|EZRegex:
    """ Generates a valid regex replacement string, using Python f-string like syntax.

        Example:
            ``` replace("named: {group}, numbered: {1}, entire: {0}") ```

        Like Python f-strings, use {{ and }} to specify { and }

        Set the `rtn_str` parameter to True to have it return an EZRegex type instead of a string

        Note: Remember that index 0 is the entire match

        There's a few of advantages to using this instead of just the regular regex replacement syntax:
        - It's consistent between dialects
        - It's closer to Python f-string syntax, which is cleaner and more familiar
        - It handles numbered, named, and entire replacement types the same
    """
    ...


def is_exactly(input:InputType) -> EZRegex:
    "This matches the string if and only if the entire string is exactly equal to `input`"
    ...

def literal(input:InputType) -> EZRegex:
    "This is a redundant function. You should always be able to use `... + 'stuff'` just as easily as `... + literal('stuff')`"
    ...

# For adding raw regex statements without sanatizing them
def raw(regex:str) -> EZRegex:
    """ If you already have some regular regex written, and you want to incorperate
        it, this will allow you to include it without sanatizing all the backslaches
        and such, which all the other EZRegexs do automatically
    """
    ...



literally_anything: EZRegex
"*Any* character, include newline"
signed: EZRegex
"a signed number, including 123, -123, and +123"
unsigned: EZRegex
"Same as number. Will not match +123"
plain_float: EZRegex
"Will match 123.45 and 123."
full_float: EZRegex
"Will match plain_float as well as things like 1.23e-10 and 1.23e+10"
int_or_float: EZRegex
ow: EZRegex
"\"Optional Whitechunk\""


ASCII: EZRegex
""" Make matching words, word boundaries, digits, and whitespace perform ASCII-only matching instead of full Unicode
matching. This is only meaningful for Unicode (str) patterns, and is ignored for bytes patterns
"""
IGNORECASE: EZRegex
""" Perform case-insensitive matching, including expressions that explicitly use uppercase members. Full Unicode matching
(such as Ü matching ü) also works unless the ASCII flag is used to disable non-ASCII matches. The current locale does not
change the effect of this flag unless the LOCALE flag is also used
"""
DOTALL: EZRegex
""" Make the '.' special character match any character at all, including a newline. It's recommended you simply use
literally_anything instead
"""
LOCALE: EZRegex
"Try not to use this, and rely on unicode matching instead"
MULTILINE: EZRegex
" This is automatically inserted when using line_start and line_end, you shouldn't need to add it manually "
UNICODE: EZRegex
" Match using the full unicode standard, instead of just ASCII. Enabled by default, and therefore redundant."



# Compile all regexs in regexs.jsonc to their appropriate languages
import re


try:
    import jstyleson
    import json
    import ezregex as ez
    import traceback

    r_special_chars_map = {
        'a': '\a',
        'b': '\b',
        't': '\t',
        'n': '\n',
        'v': '\v',
        'f': '\f',
        'r': '\r',
    }
    r_special_chars_regex = re.compile(r'\\([abtnvfr])')

    langs = {'py': ez.python, 'js': ez.javascript, 'r': ez.R, 'pcre2': ez.PCRE2}

    with open('data/regexs.jsonc') as f:
        regexs = jstyleson.load(f)

    # Compile regexs
    output = {
        'py': [],
        'js': [],
        'r': [],
        'pcre2': []
    }
    for i in regexs:
        for lang in langs:
            if 'worksIn' in i and lang not in i['worksIn']:
                continue
            if 'doesntWorkIn' in i and lang in i['doesntWorkIn']:
                continue

            try:
                compiled = str(eval(i['re'], langs[lang].__dict__))
            except Exception as e:
                print(f'Error compiling regex `{i["re"]}` in {lang}: {str(e)}\n{traceback.format_exc()}')
                exit(1)

            # This is because the REZRegex class uses doubles backslashes. This is good, when copying to
            # source code (our target use case), but when stuffed directly into a file, and read twice
            # (our tests compilation pipeline) it breaks
            if lang == 'r':
                compiled = compiled.replace('\\\\', '\\')
                compiled = re.sub(r_special_chars_regex, lambda m: r_special_chars_map[m.group(1)], compiled)

            output[lang].append({
                'ezregex': i['re'],
                'regex': compiled,
                'should': i['should'],
                'shouldnt': i['shouldnt'],
            })

    with open('data/compiled_regexs.json', 'w') as f:
        json.dump(output, f)

    # Compile replacements
    with open('data/replacements.jsonc') as f:
        replacements = jstyleson.load(f)

    output = {
        'py': [],
        'js': [],
        'r': [],
        'pcre2': []
    }
    for i in replacements:
        for lang in langs:
            if 'worksIn' in i and lang not in i['worksIn']:
                continue
            if 'doesntWorkIn' in i and lang in i['doesntWorkIn']:
                continue

            try:
                compiled = str(eval(i['re'], langs[lang].__dict__))
                compiled_repl = str(eval(i['repl'], langs[lang].__dict__))
            except Exception as e:
                print(f'Error compiling regex `{i["re"]}` in {lang}: {str(e)}\n{traceback.format_exc()}')
                exit(1)

            if lang == 'r':
                compiled = compiled.replace('\\\\', '\\')
                compiled_repl = compiled_repl.replace('\\\\', '\\')
                compiled = re.sub(r_special_chars_regex, lambda m: r_special_chars_map[m.group(1)], compiled)
                compiled_repl = re.sub(r_special_chars_regex, lambda m: r_special_chars_map[m.group(1)], compiled_repl)

            output[lang].append({
                'ezregex': i['re'],
                'regex': compiled,
                'ezrepl': i['repl'],
                'repl': compiled_repl,
                'base': i['base'],
                'after': i['after'],
            })

    with open('data/compiled_replacements.json', 'w') as f:
        json.dump(output, f)

except Exception as e:
    import sys
    print(f'Error in the compile_regexs.py script: {str(e)}', file=sys.stderr)
    raise e


import re

import jstyleson

from ezregex import *
from ezregex import python


def test_replace_func():
    assert replace('|{g}|{1}|{0}|') == '|' + rgroup('g') + '|' + rgroup(1) + '|' + replace_entire + '|'
    assert replace("{group}this is am{group}mtest") == rgroup('group') + 'this is am' + rgroup('group') + "mtest"
    assert replace("this is {{ not a thing") == "this is { not a thing"
    assert replace("also not }} a thing") == "also not } a thing"
    assert replace("still }}not{{ a thing") == "still }not{ a thing"
    assert replace("also {{not}} a thing") == "also {not} a thing"
    assert replace("but {group} is and {1} is") == "but " + rgroup('group') + " is and " + rgroup(1) + " is"
    assert replace("{group}{g}") == rgroup('group') + rgroup('g')



# pyright: reportArgumentType = false
# pyright: reportUndefinedVariable = false
from ..base import load_base
from ..EZRegex import EZRegex
from .PCRE2EZRegex import PCRE2EZRegex

globals().update(load_base(PCRE2EZRegex, lambda num_or_name, cur=...: fr'{cur}${{{num_or_name}}}'))

# we want *lowercase* \z, not \Z
is_exactly = isExactly = lambda s: fr'\A{s}\z'


import ast
import os
from copy import deepcopy
from pathlib import Path

# from clipboard import copy
from rich import print

# ASSUMPTION: Groups are designated using the form "Group: <name>\n<optional description>"
# ASSUMPTION: strings below variables act as the descriptions for those variables
# ASSUMPTION: There aren't any extraneous variables or functions in the .pyi dialect files
# TODO: This needs to add to groups instead of setting groups, so if we specify a group in a dialect, it doesn't reset
# the docs in that group to just the ones specified in the dialect, it also includes the base ones
class DocGenerator(ast.NodeVisitor):
    """ This parses the .pyi file and gets all the relevant info out of it """
    def __init__(self, node) -> None:
        self.docs = {}
        self.group = None
        self.key = None

        # ASSUMPTION: all dialects inhereit the base dialect, and only the base dialect
        with open('./base/interface.py') as f:
                self.visit(ast.parse(f.read(), type_comments=True))
        self.visit(node)

    def visit_AnnAssign(self, node: ast.AnnAssign):
        self.key = node.target.id # type: ignore
        self.docs[self.group][self.key] = [None, '']

    def visit_Expr(self, node: ast.Expr):
        if type(node.value) is ast.Constant:
            if type(node.value.value) is str and node.value.value.strip().startswith('Group: '):
                self.group = node.value.value.strip().splitlines()[0].strip()[len('Group: '):]
                self.docs[self.group] = {'description': '\n'.join(node.value.value.strip().splitlines()[1:]).strip().replace('\n', '')}

            elif self.key:
                self.docs[self.group][self.key][1] = node.value.value.strip().replace('\n', '')
                self.key = None

    def visit_FunctionDef(self, node: ast.FunctionDef):
        n = deepcopy(node)
        n.body = []
        self.docs[self.group][node.name] = [ast.unparse(n)[4:-1], ast.get_docstring(node)]

    def delete_key_recursive(self, dictionary, key_to_delete):
        keys_to_delete = []
        for key, value in dictionary.items():
            if isinstance(value, dict):
                self.delete_key_recursive(value, key_to_delete)
            elif key == key_to_delete:
                keys_to_delete.append(key)
        for key in keys_to_delete:
            del dictionary[key]

    # ASSUMPTION: The only way we remove parts from the base dialect is using the del keyword
    def visit_Delete(self, node: ast.Delete):
        for i in node.targets:
            for group, d in self.docs.items():
                if i.id in d:
                    del d[i.id]

    # def visit_Module(self, node: ast.Module):
        # if node.body[0].module == 'base.interface' and node.body[0].names[0].name =='*':
        #     with open('./base/interface.py') as f:
        #         self.visit(ast.parse(f.read(), type_comments=True))

    # def generic_visit(self, node):
        # print(type(node).__name__)
        # super().generic_visit(node)


# Go through all the .pyi files and generate the docs for each dialect
docs = {}
for path in os.walk('.'):
    for file in path[2]:
        # ASSUMPTION: we don't want to parse api.pyi, and it's the only *.pyi file we *don't* want to parse
        if file.endswith('.pyi') and not file.endswith(('api.pyi', 'EZRegex.pyi')):
            p = os.path.join(path[0] + '/' + file)
            print('Parsing', p)
            with open(p) as f:
                s = f.read()
                # ASSUMPTION: the official name of all the dialects is the same as the folder their in
                docs[path[0][2:]] = DocGenerator(ast.parse(s, type_comments=True)).docs

# TODO: This should probably move somewhere else eventually
operator_docs = {
    "`+`, `<<`, `>>`": "These all do the same thing: combine expressions",
    "`*`": "Multiplies an expression a number of times. `expr * 3` is equivelent to `expr + expr + expr`. Can also be used like `expr * ...` is equivalent to `anyAmt(expr)`",
    "`+`": "A unary + operator acts exactly as a match_max() does, or, if you're familiar with regex syntax, the + operator",
    "`[]`": "expr[2, 3] is equivalent to `match_range(2, 3, expr)`\n\t- expr[2, ...] or expr[2,] is equivalent to `at_least(2, expr)`\n\t- expr[... , 2] is equivalent to `at_most(2, expr)`\n\t- expr[...] or expr[0, ...] is equivelent to `at_least_0(expr)`\n\t- expr[1, ...] is equivalent to `at_least_1(expr)`",
    "`&`": "Coming soon! This will work like the + operator, but they can be out of order. Like an and operation.",
    "`|`": "Coming soon! This will work like an or operation, which will work just like anyOf()",
}


if __name__ == '__main__':
    sdocs = ''
    for dialect, spec in docs.items():
        s = ''
        # Iterate through the groups
        for group, elements in spec.items():
            #  style="padding-left: 20px;"
            s += f'<details>\n\t<summary>{group.title()}</summary>\n\n'
            # Add the group description, if there is one
            if 'description' in elements:
                s += '#### ' + elements.pop('description') + '\n'

            # Iterate through the elements within the groups
            for element, about in elements.items():
                signature, description = about

                s += '- '
                # if signature:
                    # Remove the last 27 chars of the functions, because those are the "-> ..." part
                    # s += str(inspect.signature(getattr(er, element)))[:-27]
                s += (signature[:-10] if signature.endswith('-> EZRegex') else signature) if signature else element

                s += '\n'

                # Add the additional explanation, if there is one
                if description:
                    s += '\t- ' + description + '\n'
            s += '\n</details>\n\n'

        sdocs += f'<details>\n\t<summary><strong><u>{dialect}</u></strong></summary>{s}</details>\n'

    # sdocs += '<details>\n\t<summary>Operators</summary>\n\n'
    # for op, desc in operator_docs.items():
    #     sdocs += '- ' + op + '\n\t- ' + desc + '\n'

        # Leave this open, cause we have the additional 2 operators manually added in the README
        # (because they're not relevant to ezregex.org)
        # s += '\n</details>\n\n'

    # The exact same thing, just using headers instead of collapsible sections
    # else:
    #     s = ''
    #     # Iterate through the groups
    #     for group, elements in __groups__.items():
    #         s += '### ' + group.title() + '\n'
    #         if group in __docs__['groups_docs']:
    #             s += '#### ' + __docs__['groups_docs'][group] + '\n'

    #         # Iterate through the elements within the groups
    #         for element in elements:
    #             s += '- ' + element
    #             if inspect.isfunction(getattr(er, element)):
    #                 # Remove the last 27 chars of the functions, because those are the "-> ..." part
    #                 s += str(inspect.signature(getattr(er, element)))[:-27]
    #             s += '\n'
    #             # Add the additional explanation, if there is one
    #             if element in __docs__ and __docs__[element] is not None:
    #                 s += '\t- ' + __docs__[element] + '\n'

        # s += '### Operators\n'
        # for op, desc in operator_docs.items():
        #     s += '- ' + op + '\n\t- ' + desc + '\n'

    # print(sdocs)
    # copy(sdocs)
    # This automatically opens the README and inserts the docs between the markdown comments.
    readme = Path(__file__).parent.parent / 'README.md'
    print(readme)
    with open(readme, 'r') as f:
        lines = f.readlines()

    start = lines.index('<!-- Start of generated docs -->\n') + 1
    end = lines.index('<!-- End of generated docs -->\n')
    # First, get rid of the old one
    lines = lines[:start] + lines[end:]
    # Now add the new one
    lines.insert(start, sdocs)

    with open(readme, 'w') as f:
        f.writelines(lines)



import pytest

import ezregex as er
from ezregex import *
from ezregex import EZRegex, python


def test_no_empty_strings():
    with pytest.raises(ValueError):
        str(group('', name='i') + optional(digit))

    with pytest.raises(ValueError):
        str(match_max(group('')))

    with pytest.raises(ValueError):
        str(amt(2, ''))

    with pytest.raises(ValueError):
        str(group('abc', name=''))


def test_any_of():
    # """ _any_of_func """
    assert er.anyof('aiLmsux', split=True, chars=True)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof('aiLmsux', split=True, chars=True)._compile(False)}'"
    # assert er.anyof('aiLmsux', split=False, chars=True)._compile(False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=True)._compile(False)}'"
    assert er.anyof('aiLmsux', split=None, chars=True)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof('aiLmsux', split=None, chars=True)._compile(False)}'"
    assert er.anyof('aiLmsux', split=True, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof('aiLmsux', split=True, chars=False)._compile(False)}'"
    # assert er.anyof('aiLmsux', split=False, chars=False)._compile(False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=False)._compile(False)}'"
    assert er.anyof('aiLmsux', split=None, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof('aiLmsux', split=None, chars=False)._compile(False)}'"
    assert er.anyof('aiLmsux', split=True, chars=None)._compile(False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof('aiLmsux', split=True, chars=None)._compile(False)}'"
    # assert er.anyof('aiLmsux', split=False, chars=None)._compile(False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=None)._compile(False)}'"
    assert er.anyof('aiLmsux', split=None, chars=None)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof('aiLmsux', split=None, chars=None)._compile(False)}'"

    # assert er.anyof(*list('aiLmsux'), split=True, chars=True)._compile(False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=True)._compile(False)}'"
    assert er.anyof(*list('aiLmsux'), split=False, chars=True)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=False, chars=True)._compile(False)}'"
    assert er.anyof(*list('aiLmsux'), split=None, chars=True)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=None, chars=True)._compile(False)}'"
    # assert er.anyof(*list('aiLmsux'), split=True, chars=False)._compile(False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=False)._compile(False)}'"
    assert er.anyof(*list('aiLmsux'), split=False, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof(*list('aiLmsux'), split=False, chars=False)._compile(False)}'"
    assert er.anyof(*list('aiLmsux'), split=None, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)', \
        f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof(*list('aiLmsux'), split=None, chars=False)._compile(False)}'"
    # assert er.anyof(*list('aiLmsux'), split=True, chars=None)._compile(False) == Error, \
    #   f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=None)._compile(False)}'"
    assert er.anyof(*list('aiLmsux'), split=False, chars=None)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=False, chars=None)._compile(False)}'"
    assert er.anyof(*list('aiLmsux'), split=None, chars=None)._compile(False) == "[aiLmsux]", \
        f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=None, chars=None)._compile(False)}'"


def test_any_char_except():
    # """ _any_char_except_func """
    assert er.any_char_except('abcd')._compile(False) == '[^abcd]'
    assert er.any_char_except(*list('abcd'))._compile(False) == '[^abcd]'



def test_misc():
    """ Just threw a bunch of stuff in here for now, I need to move it elsewhere eventually """
    a = word + ow
    # b = stuff + UNICODE
    c = IGNORECASE + '9'
    assert a + c == word + ow + IGNORECASE + '9', f"{a + c} != {word + ow + IGNORECASE + '9'}"

    a = str(PythonEZRegex(r'\s+'))
    with pytest.raises(TypeError):
        b = str(PythonEZRegex(raw(r'\s+')))
    c = r'\s+'
    d = str(raw(r'\s+'))
    # e = str(whitespace + matchMax)
    assert a == c == d, f'\na: {a}\n c: {c}\n d: {d}\n e: {e}'
    # assert (word + ow + anything + ':').test('word    d:', show=False)
    # assert not (word + ow + anything + ':').test('word', show=False)
    assert 'word    d:' in (word + ow + anything + ':')

    test = word + chunk
    test += word
    assert str(test) == str(word + chunk + word), f"{str(test)} != {str(word + chunk + word)}"
    assert test == word + chunk + word
    assert either('(' + word + ')', '.') == either(er.literal('(') + word() + er.literal(')'), '.'), f"{either('(' + word + ')', '.')} != {either(er.literal('(') + word() + er.literal(')'), '.')}"
    assert str(er.ifFollowedBy(word)) == r'(?=\w+)'

    #TODO: assert (word + ow + anything + ':') in 'word    d:'
    #TODO: assert (word + ow + anything + ':') not in 'word'
    assert str(word) == r'\w+', f'{word}'



import json
from logging import warning
from sys import version_info

if version_info < (3, 12):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

from warnings import warn

import jstyleson
from pydantic import TypeAdapter, ValidationError

# from ezregex.api import APIStructure
from ezregex import *
from ezregex import api, python

# from typing_extensions import TypedDict # Required by pydantic for python < 3.12
# import importlib, sys

# def import_stub(stubs_path, module_name):
#     sys.path_hooks.insert(0,
#         importlib.machinery.FileFinder.path_hook(
#             (importlib.machinery.SourceFileLoader, ['.pyi']))
#     )
#     sys.path.insert(0, stubs_path)

#     try:
#         return importlib.import_module(module_name)
#     finally:
#         sys.path.pop(0)
#         sys.path_hooks.pop(0)


# import importlib.util
# import sys

# file = "/home/anastasia/hello/python/ezregex/ezregex/api.pyi"
# sys.path_hooks.insert(0,
#     importlib.machinery.FileFinder.path_hook(
#         (importlib.machinery.SourceFileLoader, ['.pyi']))
# )
# sys.path.insert(0, file)
# spec = importlib.util.spec_from_file_location("module.name", file)
# api = importlib.util.module_from_spec(spec)
# sys.modules["module.name"] = api
# spec.loader.exec_module(api)
# print(api)

# def import_stub(stub_path, module_name):
#     spec = importlib.util.spec_from_file_location(module_name, stub_path)
#     module = importlib.util.module_from_spec(spec)
#     spec.loader.exec_module(module)
#     return module

# import_stub("/home/anastasia/hello/python/ezregex/ezregex/api.pyi", 'api')

# print(api_types)


# This is just copied from api.pyi, because I'm tired of trying to figure out how to import .pyi files for the moment
Group = TypedDict(
    'Group',
    {
        'string': str,
        'end': int,
        'start': int,
        'color': str
    }
)

Match = TypedDict(
    'Match',
    {
        'string': str,
        'string HTML': str,
        'parts': list[list[str|None]],
        'end': int,
        'start': int,
        'color': str
    }
)

Matches = TypedDict(
    'Matches',
    {
        'match': Match,
        'unnamed groups': dict[int, Group],
        'named groups': dict[str, Group],
    }
)
APIStructure = TypedDict(
    "APIStructure",
    {
        'regex': str,
        'string': str,
        'string HTML': str,
        'parts': list[list[str|None]],
        'matches': list[Matches],
        'replaced': str | None,
        'split': list[str|None] | None,
    }
)

# TODO: test that a pattern like <span> does not break things
def test_correct_output():
    with open('data/regexs.jsonc') as f:
        regexs = jstyleson.load(f)

    for i in regexs:
        regex_str = i['re']
        match = i['should']
        dontmatch = i['shouldnt']
        if 'worksIn' in i and 'py' not in i['worksIn']:
            continue
        if 'doesntWorkIn' in i and 'py' in i['doesntWorkIn']:
            continue

        try:
            regex = eval(regex_str, python.__dict__)
        except Exception as err:
            raise AssertionError(f"Failed to parse pattern `{regex_str}`") from err
        try:
            resp = api(regex)
        except NotImplementedError as err:
            warning(err)
        except Exception as err:
            raise AssertionError(f"Failed on pattern `{regex_str}` -> `{regex}`") from err
        try:
            TypeAdapter(APIStructure).validate_python(resp)
        except ValidationError as err:
            raise AssertionError(f"Invalid schema from {regex_str}:\n{json.dumps(resp, indent=4)}\n{'-'*20}\nErrors:\n{err.errors()}\n{'-'*20}\n") from err



from ..EZRegex import EZRegex


class PCRE2EZRegex(EZRegex):
    _escape_chars=b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f'
    _repl_escape_chars=b'$'

    def _final_func(self, s:str) -> str:
        if self.replacement:
            # This is how you escape a $ in a replacement string
            return s.replace(r'\$', '$$')
        return s


from ..EZRegex import EZRegex
from functools import partial

class JavaScriptEZRegex(EZRegex):
    _escape_chars=b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f/'
    _repl_escape_chars=b''

    def __init__(self, *args, string_anchor_used=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__setattr__('string_anchor_used', string_anchor_used, True)

    # These 3 methods are copied and modified from EZRegex
    # TODO: I'm sure there's a cleaner way to do this
    def _copy(self, definition=..., sanatize=..., replacement=..., flags=..., string_anchor_used=...):
        if definition is Ellipsis:
            definition = self._compile()
        if sanatize is Ellipsis:
            sanatize = self._sanatize
        if replacement is Ellipsis:
            replacement = self.replacement
        if flags is Ellipsis:
            flags = self._flags
        if string_anchor_used is Ellipsis:
            string_anchor_used = self.string_anchor_used

        return type(self)(definition, sanatize=sanatize, replacement=replacement, flags=flags, string_anchor_used=string_anchor_used)

    def __add__(self, thing):
        return self._copy(
            self._funcList + [partial(lambda cur=...: cur + self._sanitizeInput(thing))],
            sanatize=self._sanatize or thing._sanatize if isinstance(thing, EZRegex) else self._sanatize,
            replacement=self.replacement or thing.replacement if isinstance(thing, EZRegex) else self.replacement,
            flags=(self._flags + thing.flags) if isinstance(thing, EZRegex) else self._flags,
            string_anchor_used=self.string_anchor_used or thing.string_anchor_used if isinstance(thing, EZRegex) else self.string_anchor_used
        )

    def __radd__(self, thing):
        return self._copy([partial(lambda cur=...: self._sanitizeInput(thing) + cur)] + self._funcList,
            sanatize=self._sanatize or thing._sanatize if isinstance(thing, EZRegex) else self._sanatize,
            replacement=self.replacement or thing.replacement if isinstance(thing, EZRegex) else self.replacement,
            flags=(self._flags + thing.flags) if isinstance(thing, EZRegex) else self._flags,
            string_anchor_used=self.string_anchor_used or thing.string_anchor_used if isinstance(thing, EZRegex) else self.string_anchor_used
        )

    def _flag_func(self, final):
        if self.string_anchor_used and 'm' in self.flags:
            raise ValueError('string_starts_with and string_ends_with don\'t work with the multiline flag')
        if self.replacement:
            return final
        else:
            return f'/{final}/{self.flags}'

    # def _final_func(self, s:str) -> str:
    #     return f'/{s}/'



"""
To tests parts of induvidual regexs that can't be covered by regexs.jsonc tests, things like throwing errors
These tests are not specific to any given dialect
This file gets run by pytest
"""

import pytest
import ezregex.javascript as js
import ezregex.R as r

def test_js_string_anchor_throws_errors():
    assert js.string_starts_with.string_anchor_used is True
    assert js.string_ends_with.string_anchor_used is True
    assert js.digit.string_anchor_used is False
    with pytest.raises(ValueError):
        (js.string_starts_with + js.MULTILINE).str()
    with pytest.raises(ValueError):
        (js.string_ends_with + js.MULTILINE).str()
    with pytest.raises(ValueError):
        (js.string_starts_with + js.line_ends_with).str()
    with pytest.raises(ValueError):
        (js.string_ends_with + js.line_starts_with).str()
    assert js.string_starts_with.str() == '/^/'
    assert js.string_ends_with.str() == '/$/'

# TODO: this test should probably be in test_misc.py instead
def test_js_flags_added_on_outside():
    assert js.digit.str() == r'/\d/'
    assert (js.digit + js.MULTILINE).str() == r'/\d/m'
    assert js.line_starts_with.str() == '/^/m'
    assert js.line_ends_with.str() == '/$/m'
    assert (js.line_ends_with + js.MULTILINE).str() == '/$/m'

def test_r_no_named_groups():
    with pytest.raises(ValueError):
        r.group('test', name='test').str()
    with pytest.raises(ValueError):
        r.group('test', 'test').str()
    with pytest.raises(ValueError):
        r.group('test', '1').str()


import logging
import re
from abc import ABC
from functools import partial
from .api import api
from .generate import *
from .invert import invert

from .base import base, psuedonymns

# TODO: Seperate EZRegex into a "bytes" mode vs "string" mode
# TODO: consider changing add_flags to "outer" or "end" or something
# TODO: Seriously consider removing the debug functions
# TODO: in all the magic functions assert that we're not mixing dialects
# TODO: figure out if theres a way to make a "change dialect" function
# TODO: B.join((A, C))
# TODO: sre.concatenate((A, B, C)) or A + B + C (if I'd fancy the plus-style syntax more than flow-style)

class EZRegex(ABC):
    """ Represent parts of the Regex syntax. Should not be instantiated by the user directly."""

    def __init__(self, definition, *, sanatize=True, replacement=False, flags=''):
        """
        The workhorse of the EZRegex library. This represents a regex pattern that can be combined
        with other EZRegexs and strings. Ideally, this should only be called internally, but it should
        still work from the user's end
        """
        # Set attributes like this so the class can remain "immutable", while still being usable
        self.__setattr__('_flags', flags, True)
        self.__setattr__('_sanatize', sanatize, True)
        self.__setattr__('replacement', replacement, True)

        if isinstance(definition, str):
            self.__setattr__('_funcList', [lambda cur=...: cur + definition], True)
        elif callable(definition):
            self.__setattr__('_funcList', [definition], True)
        elif isinstance(definition, list):
            self.__setattr__('_funcList', definition, True)

    # Private functions
    def _flag_func(self, final:str) -> str:
        """ This function is called to add the flags to the regex. It gets called even if there are no flags """
        if self.flags and not self.replacement:
            return f'(?{self.flags}){final}'
        return final


    def _final_func(self, s:str) -> str:
        return s

    def _escape(self, pattern:str):
        return self.escape(pattern, self.replacement)

    @classmethod
    def escape(cls, pattern:str, replacement=False):
        """ Available as a class method, so we can escape strings when defining singletons
            The user could use this, but I can't think of a reason they would want to.
        """
        # This function was modified from the one in /usr/lib64/python3.12/re/__init__.py line 255
        _special_chars_map = {i: '\\' + chr(i) for i in (cls._repl_escape_chars if replacement else cls._escape_chars)}
        return pattern.translate(_special_chars_map)

    def _sanitizeInput(self, i, add_flags=False):
        """ Instead of raising an error if passed a strange datatype, it now tries to cast it to a string """
        # If it's another chain, compile it
        if isinstance(i, EZRegex):
            return i._compile(add_flags=add_flags)
        # It's a string (so we need to escape it)
        # If this is a replacement string, it will automatically escape based on _repl_escape_chars
        elif isinstance(i, str):
            return self._escape(i)
        # A couple of singletons use bools and None as kwargs, just ignore them and move on
        elif i is None or isinstance(i, bool):
            return i
        # A couple singletons use ints as input, just cast it to a string and don't throw a warning
        elif isinstance(i, int):
            return str(i)
        # It's something we don't know, but try to cast it to a string anyway
        else:
            try:
                logging.warning(f"Type {type(i)} passed to EZRegex, auto-casting to a string. Special characters will will not be escaped.")
                return str(i)
            except:
                raise ValueError(f'Incorrect type {type(i)} given to EZRegex parameter: Must be string or another EZRegex chain.')

    def _compile(self, add_flags=True):
        regex = ''
        for func in self._funcList:
            regex = func(cur=regex) # type: ignore

        # Add the flags
        if add_flags:
            # Remove duplicate flags
            self.__setattr__('_flags', ''.join(set(self._flags)), True)

            regex = self._flag_func(regex)

            # This has to go in the add_flags scope so it only runs at the very end, like flags
            regex = self._final_func(regex)
        return regex

    def _copy(self, definition=..., sanatize=..., replacement=..., flags=...):
        if definition is Ellipsis:
            definition = self._compile()
        if sanatize is Ellipsis:
            sanatize = self._sanatize
        if replacement is Ellipsis:
            replacement = self.replacement
        if flags is Ellipsis:
            flags = self._flags

        return type(self)(definition, sanatize=sanatize, replacement=replacement, flags=flags)

    def _base(self, element, /, *args, **kwargs):
        """ Constructs the base element specified, and returns it passed with any additional arguements """
        return type(self)(**base[element])(*args, **kwargs)

    # Regular functions
    def str(self):
        return self.__str__()

    def debug(self):
        try:
            from Cope import debug
        except ImportError:
            print(f"Compiled ezregex string = {self}")
        else:
            debug(self, name='Compiled ezregex string', calls=2)
        return self

    def copy(self, add_flags=True):
        try:
            from clipboard import copy  # type: ignore
        except ImportError as err:
            raise ImportError(
                'Please install the clipboard module in order to auto copy ezregex expressions (i.e. pip install clipboard)'
            ) from err
        else:
            copy(self._compile(add_flags=add_flags))

    def test(self, testString=None, show=True, context=True) -> bool:
        """ Tests the current regex expression to see if it's in @param testString.
            Returns the match objects (None if there was no match)
        """
        try:
            from rich import print as rprint
            from rich.panel import Panel
            from rich.text import Text
        except ImportError:
            raise ImportError("The rich library is required to use the EZRegex.test() method. Try running `pip install rich`")

        # json = self._matchJSON(testString=testString)
        json = api(self, test_string=testString)
        if not show:
            return bool(len(json['matches']))

        st = Text()  # String
        gt = Text()  # Groups (all the group-related text)
        defaultColor = 'bold'
        textColor = ''

        st.append("Testing expression", style=defaultColor)
        # Add the context line
        # st.append(f' (from {get_context(get_metadata(2), False, True, True).strip()})', style=defaultColor)
        st.append(':\n', style=defaultColor)

        # The expression we're testing
        st.append(f'\t{json["regex"]}\n', style=textColor)
        st.append("for matches in:\n\t", style=defaultColor)

        # Add the main string
        for color, background, part in json['parts']:
            st.append(part, style=color if background is None else f'{color} on {background}')
        st.append('\n')

        # Add all the matches and groups
        for m in json['matches']:
            gt.append('Match = "')
            for color, background, part in m['match']['parts']:
                gt.append(part, style=color if background is None else f'{color} on {background}')
            gt.append('" ')
            gt.append(f"({m['match']['start']}:{m['match']['end']})", style='italic bright_black')
            gt.append('\n')
            if len(m['unnamed groups']):
                gt.append('Unnamed Groups:\n')
            for id, group in m['unnamed groups'].items():
                gt.append(f'\t{id}: "')
                gt.append(['string'], style=group['color'])
                gt.append('" ')
                gt.append(f"({group['start']}:{group['end']})", style='italic bright_black')
                gt.append('\n')
            if len(m['named groups']):
                gt.append('Named Groups:\n')
            for name, group in m['named groups'].items():
                gt.append(f'\t{name}: "')
                gt.append(group['string'], style=group['color'])
                gt.append('" ')
                gt.append(f"({group['start']}:{group['end']})", style='italic bright_black')
                gt.append('\n')
            gt.append('\n')

        # Assemble everything into a panel
        rprint(Panel(Text.assemble(*st, '\n', *gt),
            title='Testing Regex',
            subtitle=Text('Found\n', style='blue') if len(json['matches'])
                else Text('Not Found\n', style='red')))  #, border_style='dim green')

        return bool(len(json['matches']))

    def inverse(self, amt=1, **kwargs):
        """ "Inverts" the current Regex expression to give an example of a string it would match.
            Useful for debugging purposes. """
        return '\n'.join([invert(self._compile(), **kwargs) for _ in range(amt)])

    def invert(self, amt=1, **kwargs):
        return self.inverse(amt, **kwargs)

    # Elemental functions
    def group(self, name=None):
        return self._base('group', self, name=name)

    def named(self, name):
        return self.group(name)

    @property
    def unnamed(self):
        return self.group()

    def if_not_preceded_by(self, input):
        return self._base('if_not_preceded_by', input) + self

    def if_preceded_by(self, input):
        return self._base('if_preceded_by', input) + self

    def if_not_proceded_by(self, input):
        return self + self._base('if_not_proceded_by', input)

    def if_proceded_by(self, input):
        return self + self._base('if_proceded_by', input)

    def if_not_followed_by(self, input):
        return self.if_not_proceded_by(input)

    def if_followed_by(self, input):
        return self.if_proceded_by(input)

    def if_enclosed_with(self, open, closed=None):
        return self._base('if_enclosed_with', self, open, closed)

    @property
    def optional(self):
        return self._base('optional', self)

    @property
    def repeat(self):
        return self._base('repeat', self)

    @property
    def exactly(self):
        return self._base('is_exactly', self)

    def at_least(self, min):
        return self._base('at_least', min, self)

    def more_than(self, min):
        return self._base('more_than', min, self)

    def amt(self, amt):
        return self._base('amt', amt, self)

    def at_most(self, max):
        return self._base('at_most', max, self)

    def between(self, min, max, greedy=True, possessive=False):
        return self._base('between', min, max, self, greedy=greedy, possessive=possessive)

    def at_least_one(self, greedy=True, possessive=False):
        return self._base('at_least_one', self, greedy=greedy, possessive=possessive)

    def at_least_none(self, greedy=True, possessive=False):
        return self._base('at_least_none', self, greedy=greedy, possessive=possessive)

    def or_(self, input):
        return self._base('either', self, input)

    @property
    def ASCII(self):
        return self.set_flags('a')

    @property
    def IGNORECASE(self):
        return self.set_flags('i')

    @property
    def DOTALL(self):
        return self.set_flags('s')

    @property
    def LOCALE(self):
        return self.set_flags('L')

    @property
    def MULTILINE(self):
        return self.set_flags('m')

    @property
    def UNICODE(self):
        return self.set_flags('u')

    # Named operator functions
    def append(self, input):
        return self + input

    def prepend(self, input):
        return input + self

    # Flag functions
    @property
    def flags(self):
        return self._flags

    def set_flags(self, to):
        return self._copy(flags=to)

    def add_flag(self, flag):
        if flag not in self._flags:
            return self._copy(flags=self._flags + flag)
        return self

    def remove_flag(self, flag):
        if flag in self._flags:
            return self._copy(flags=self._flags.replace(flag, ''))
        return self

    # Magic Functions
    def __call__(self, *args, **kwargs):
        """ This should be called by the user to specify the specific parameters of this instance i.e. anyof('a', 'b') """
        # If this is being called without parameters, just compile the chain.
        # If it's being called *with* parameters, then it better be a fundemental
        # member, otherwise that doesn't make any sense.
        if len(self._funcList) > 1:
            if not len(args) and not len(kwargs):
                return self._compile()
            else:
                raise TypeError("You're trying to pass parameters to a chain of expressions. That doesn't make any sense. Stop that.")

        # Sanatize the arguments
        if self._sanatize:
            args = list(map(self._sanitizeInput, args))

        _kwargs = {}
        for key, val in kwargs.items():
            _kwargs[key] = self._sanitizeInput(val) if self._sanatize else val

        return self._copy([partial(self._funcList[0], *args, **_kwargs)])

    def __str__(self):
        return self._compile()

    def __repr__(self):
        return f'{type(self).__name__}("{self}")'

    def __eq__(self, thing):
        """ NOTE: This will return True for equivelent EZRegex expressions of different dialects """
        return self._sanitizeInput(thing, add_flags=True) == self._compile()

    def __mul__(self, amt):
        if amt is Ellipsis:
            return self._copy(f'(?{self})*', sanatize=False)
        rtn = self
        # This isn't optimal, but it's unlikely anyone will use this with large numbers
        for i in range(amt-1):
            # This doesn't work
            # rtn += self
            # But this does??
            rtn = rtn + self
        return rtn

    def __rmul__(self, amt):
        return self * amt

    def __imul__(self, amt):
        return self * amt

    def __add__(self, thing):
        return self._copy(
            self._funcList + [partial(lambda cur=...: cur + self._sanitizeInput(thing))],
            sanatize=self._sanatize or thing._sanatize if isinstance(thing, EZRegex) else self._sanatize,
            replacement=self.replacement or thing.replacement if isinstance(thing, EZRegex) else self.replacement,
            flags=(self._flags + thing.flags) if isinstance(thing, EZRegex) else self._flags
        )

    def __radd__(self, thing):
        return self._copy([partial(lambda cur=...: self._sanitizeInput(thing) + cur)] + self._funcList,
            sanatize=self._sanatize or thing._sanatize if isinstance(thing, EZRegex) else self._sanatize,
            replacement=self.replacement or thing.replacement if isinstance(thing, EZRegex) else self.replacement,
            flags=(self._flags + thing.flags) if isinstance(thing, EZRegex) else self._flags
        )

    def __iadd__(self, thing):
        return self + thing

    def __and__(self, thing):
        raise NotImplementedError
        logging.warning('The & operator is unstable still. Use each() instead.')
        return EZRegex(fr'(?={self}){thing}', self.dialect, sanatize=False)

    def __rand__(self, thing):
        raise NotImplementedError
        logging.warning('The & operator is unstable still. Use each() instead.')
        return EZRegex(fr'(?={thing}){self}', self.dialect, sanatize=False)
    # The shift operators just shadow the add operators
    def __lshift__(self, thing):
        return self.__add__(thing)

    def __rlshift__(self, thing):
        return self.__radd__(thing)

    def __ilshift__(self, thing):
        return self.__iadd__(thing)
    # I don't think right and left shifts should be any different, right?
    def __rshift__(self, thing):
        return self.__add__(thing)

    def __rrshift__(self, thing):
        return self.__radd__(thing)

    def __irshift__(self, thing):
        return self.__iadd__(thing)

    def __invert__(self):
        return self.invert()

    def __pos__(self):
        comp = self._compile(add_flags=False)
        return self._copy(('' if not len(comp) else r'(?:' + comp + r')') + r'+', sanatize=False)

    def __ror__(self, thing):
        return self._copy(f'(?:{self._sanitizeInput(thing)}|{self._compile(add_flags=False)})', sanatize=False)

    def __or__(self, thing):
        logging.warning('The or operator is unstable and likely to fail, if used more than twice. Use anyof() instead, for now.')
        return self._copy(f'(?:{self._compile(add_flags=False)}|{self._sanitizeInput(thing)})', sanatize=False)

    def __xor__(self, thing):
        return NotImplementedError

    def __rxor__(self, thing):
        return NotImplementedError

    def __mod__(self, other):
        """ I would prefer __rmod__(), but it doesn't work on strings, since __mod__() is already specified for string formmating. """
        # I don't need to check this, re will do it for me
        # if not isisntance(other, str):
            # raise TypeError(f"Can't search type {type(other)} ")
        return re.search(other, self._compile(add_flags=False))

    def __hash__(self):
        if len(self._funcList) > 1:
            return hash(self._compile())
        # If we only have 1 function lined up, that means we haven't
        # been called at all. And that means we're one of the basic singletons,
        # because users aren't supposed to instantiate this class directly.
        # THAT means we can use this instance's pointer as a unique identifier.
        else:
            return hash(id(self))

    def __contains__(self, thing):
        # assert isinstance(thing, str), "`in` statement can only be used with a string"
        return re.search(self._compile(), thing) is not None

    def __rcontains__(self, thing):
        """ I guess this isn't a thing? But it really should be. """
        assert isinstance(thing, str), "`in` statement can only be used with a string"
        return re.search(self._compile(), thing) is not None

    def __getitem__(self, args):
        # digit[2, 3]    # (2, 3)
        # digit[2, ...]  # (2, Ellipsis)
        # digit[2, None] # (2, None)
        # digit[2, ]     # (2,)
        # digit[..., 3]  # (Ellipsis, 3)
        # digit[None, 3] # (None, 3)
        # digit[...:3]   # slice(Ellipsis, 3, None)
        # digit[None:3]  # slice(None, 3, None)
        # digit[:3]      # slice(None, 3, None)
        # digit[:]       # slice(None, None, None)
        # digit[2]       # 2

        # assert digit[2, 3] == match_range(2, 3, digit)
        # assert digit[2, ...] == digit[2,] == digit[2, None] == digit[2] == match_at_least(2, digit)
        # # assert digit[..., 2] == digit[0, 2] == digit[None, 2] == match_at_most(2, digit)
        # assert digit[...] == digit[0, ...] == digit[None] == at_least_0(digit)
        # assert digit[1, ...] == digit[1] == digit[1,] == digit[1, None] == at_least_1(digit)

        if type(args) is slice:
            # expr[...:end_expr] is equivalent to ZeroOrMore(expr, stop_on=end_expr)
            # assert digit[...:'foo'] == digit[None:'foo'] == digit[,'foo'] ==
            pass
        elif type(args) is not tuple or len(args) == 1:
            if type(args) is tuple:
                args = args[0]
            if args is None or args is Ellipsis or args == 0:
                # at_least_0(self)
                return self._copy(fr'(?:{self._compile(False)})*', sanatize=False)
            elif args == 1:
                # at_least_1(self)
                return self._copy(fr'(?:{self._compile(False)})+', sanatize=False)
            else:
                # match_at_least(args, self)
                return self._copy(fr'(?:{self._compile(False)}){{{args},}}', sanatize=False)
        else:
            start, end = args
            if start is None or start is Ellipsis:
                # match_at_most(2, self)
                return self._copy(fr'(?:{self._compile(False)}){{0,{end}}}', sanatize=False)
            elif end is None or end is Ellipsis:
                if start == 0:
                    # at_least_0(self)
                    return self._copy(fr'(?:{self._compile(False)})*', sanatize=False)
                elif start == 1:
                    # at_least_1(self)
                    return self._copy(fr'(?:{self._compile(False)})+', sanatize=False)
                else:
                    # match_at_least(start, self)
                    return self._copy(fr'(?:{self._compile(False)}){{{start},}}', sanatize=False)
            else:
                # match_range(start, end, self)
                return self._copy(fr'(?:{self._compile(False)}){{{start},{end}}}', sanatize=False)

    def __reversed__(self):
        return self.inverse()

    def __rich__(self):
        return self._compile()

    def __pretty__(self):
        return self._compile()

    def __setattr__(self, name, value, ignore=False):
        if ignore:
            self.__dict__[name] = value
        else:
            raise TypeError('EZRegex objects are immutable')

    def __delattr__(self, name, ignore=False):
        if ignore:
            del self.__dict__[name]
        else:
            raise TypeError('EZRegex objects are immutable')



import re
from timeit import repeat

import pytest

import ezregex as er
from ezregex import *


def test_basic():
    assert literal('test') == 'test'

def test_basic_concat():
    assert str(literal('test') + digit) == r'test\d'
    assert str('test' + digit) == r'test\d'

def test_access_dialect():
    assert type(literal('thing')) is PythonEZRegex

def test_psuedonyms():
    assert er.matchMax(digit) == er.match_max(digit)
    assert matchMax(digit) == match_max(digit)


def test_immutability():
    with pytest.raises(TypeError):
        digit.flags = 'ab'
    with pytest.raises(TypeError):
        digit.compile = lambda: ...
    with pytest.raises(TypeError):
        del digit.compile
    with pytest.raises(TypeError):
        del digit.flags

def test_test_method():
    return
    # ow = optional(whitechunk)
    params = er.group(er.atLeastNone(er.ow + er.word + er.ow + er.optional(',') + er.ow))
    function = er.word + er.ow + '(' + params + ')'
    function.test('this should match func(param1, param2 ), foo(), and bar( foo,)')

    r = 'group 1' + ':' + ow + group('stuff') + ' | ' + 'group ' + number + ': ' + group('things') + ' | ' + 'named group "' + word + '": '  + named_group('foo', 'bar')
    s = 'random stuff! and then group 1: stuff | group 2: things | named group "foo": bar  \t oh and then more random stuff'
    r.test(s)

    s = 'word1 word2 word3'
    word.test(s)

    (word + whitechunk + group('func') + ':' + group(anyof('8', '7')), 'test').test()

    # This is actually accurate, if you think about it.
    # ifFollowedBy(word).test("literal(hllow) + isExactly('thing')")# fails in _matchJSON()

    ('(' + +(anything + optional(group(comma))) + ')').test()# -- empty groups print as None

    group(+group(number) + group(anyof('98'))).test('999')

def test_no_parameters_to_chains():
    # This is sort of conceptually impossible, ish
    # with pytest.raises(TypeError):
    #     digit(6)
    # with pytest.raises(TypeError):
    #     digit(input=6)
    with pytest.raises(TypeError):
        (digit + word)(6)
    with pytest.raises(TypeError):
        (digit + word)(input=6)
    # This is also conceptually impossible without a moderate refactor
    # with pytest.raises(TypeError):
    #     match_amt(6, digit)(6)
    # with pytest.raises(TypeError):
    #     match_amt(6, digit)(input=6)

    assert match_amt(6, digit)()
    assert (digit + word)()
    assert digit()

def test_re_shadow_funcs():
    s = r'\d(\w+)'
    string = 'timmy is 6years old'
    repl = replace('number {1}')

    def eq(a, b):
        if not ((a is None) == (b is None)) and a.span() == b.span() and a.groups() == b.groups():
            raise AssertionError(f'{a} != {b}')

    eq((digit + group(word)).search(string),      re.compile(s).search(string))
    eq((digit + group(word)).match(string),       re.compile(s).match(string))
    eq((digit + group(word)).fullmatch(string),   re.compile(s).fullmatch(string))
    eq((digit + group(word)).split(string),       re.compile(s).split(string))
    eq((digit + group(word)).findall(string),     re.compile(s).findall(string))
    eq((digit + group(word)).finditer(string),    re.compile(s).finditer(string))
    eq((digit + group(word)).sub(repl, string),   re.compile(s).sub(repl, string))
    eq((digit + group(word)).subn(repl, string),  re.compile(s).subn(repl, string))

def test_flag_methods():
    assert (digit + ASCII).flags == 'a'
    with pytest.raises(TypeError):
        digit.flags = 'L'
    assert digit.flags == ''
    a = digit.set_flags('asL')
    assert a.flags == 'asL'
    assert (digit + ASCII).add_flag('L').flags == 'aL'
    assert (digit + ASCII).remove_flag('a').flags == ''
    assert (digit + ASCII).remove_flag('L').flags == 'a'

def test_elemental_methods():
    input = word
    min = 1
    max = 3
    try:
        assert digit.group() == group(digit) == digit.unnamed
        assert digit.group('test') == group(digit, 'test') == digit.named('test')
        assert digit.if_not_preceded_by(input) == if_not_preceded_by(input) + digit
        assert digit.if_preceded_by(input) == if_preceded_by(input) + digit
        assert digit.if_not_proceded_by(input) == digit + if_not_proceded_by(input)
        assert digit.if_proceded_by(input) == digit + if_proceded_by(input)
        assert digit.if_enclosed_with('|') == if_enclosed_with(digit, '|')
        assert digit.if_enclosed_with('(', ')') == if_enclosed_with(digit, '(', ')')
        assert digit.at_least(min) == at_least(min, digit)
        assert digit.more_than(min) == more_than(min, digit)
        assert digit.amt(2) == match_num(2, digit)
        assert digit.at_most(max) == at_most(max, digit)
        assert digit.between(min, max) == between(min, max, digit)
        assert digit.between(min, max, greedy=False) == between(min, max, digit, greedy=False)
        assert digit.between(min, max, possessive=True) == between(min, max, digit, possessive=True)
        assert digit.at_least_one() == at_least_one(digit)
        assert digit.at_least_one(greedy=False) == at_least_one(digit, greedy=False)
        assert digit.at_least_one(possessive=True) == at_least_one(digit, possessive=True)
        assert digit.at_least_none() == at_least_none(digit)
        assert digit.at_least_none(greedy=False) == at_least_none(digit, greedy=False)
        assert digit.at_least_none(possessive=True) == at_least_none(digit, possessive=True)
        assert digit.or_(input) == either(digit, input)

        assert MULTILINE + digit.group() == group(digit) + MULTILINE
        assert MULTILINE + digit.group('test') == group(digit, 'test') + MULTILINE
        assert MULTILINE + digit.if_not_preceded_by(input) == if_not_preceded_by(input) + MULTILINE + digit
        assert MULTILINE + digit.if_preceded_by(input) == if_preceded_by(input) + MULTILINE + digit
        assert MULTILINE + digit.if_not_proceded_by(input) == digit + if_not_proceded_by(input) + MULTILINE
        assert MULTILINE + digit.if_proceded_by(input) == digit + if_proceded_by(input) + MULTILINE
        assert MULTILINE + digit.if_enclosed_with('|') == if_enclosed_with(digit, '|') + MULTILINE
        assert MULTILINE + digit.if_enclosed_with('(', ')') == if_enclosed_with(digit, '(', ')') + MULTILINE
        assert MULTILINE + digit.at_least(min) == at_least(min, digit) + MULTILINE
        assert MULTILINE + digit.more_than(min) == more_than(min, digit) + MULTILINE
        assert MULTILINE + digit.amt(2) == match_num(2, digit) + MULTILINE
        assert MULTILINE + digit.at_most(max) == at_most(max, digit) + MULTILINE
        assert MULTILINE + digit.between(min, max) == between(min, max, digit) + MULTILINE
        assert MULTILINE + digit.between(min, max, greedy=False) == between(min, max, digit, greedy=False) + MULTILINE
        assert MULTILINE + digit.between(min, max, possessive=True) == between(min, max, digit, possessive=True) + MULTILINE
        assert MULTILINE + digit.at_least_one() == at_least_one(digit) + MULTILINE
        assert MULTILINE + digit.at_least_one(greedy=False) == at_least_one(digit, greedy=False) + MULTILINE
        assert MULTILINE + digit.at_least_one(possessive=True) == at_least_one(digit, possessive=True) + MULTILINE
        assert MULTILINE + digit.at_least_none() == at_least_none(digit) + MULTILINE
        assert MULTILINE + digit.at_least_none(greedy=False) == at_least_none(digit, greedy=False) + MULTILINE
        assert MULTILINE + digit.at_least_none(possessive=True) == at_least_none(digit, possessive=True) + MULTILINE
        assert MULTILINE + digit.or_(input) == either(digit, input) + MULTILINE
    # I usually run tests in Python3.12, so I'm just gonna disable all these tests for Python3.10 (since I have automated
    # tests run in Python3.10, and 3.10 doesn't support possessive or greedy regex operators)
    except Exception as err:
        if 'require at least Python3.11' not in str(err):
            raise err

    assert digit.optional == optional(digit)
    assert digit.repeat == repeat(digit)
    assert digit.exactly == is_exactly(digit)

    assert digit.ASCII == digit + ASCII
    assert digit.IGNORECASE == digit + IGNORECASE
    assert digit.DOTALL == digit + DOTALL
    assert digit.LOCALE == digit + LOCALE
    assert digit.MULTILINE == digit + MULTILINE

    assert digit.append(input) == digit + input
    assert digit.prepend(input) == input + digit

    assert digit.append(input).ASCII == digit + ASCII + input
    assert digit.prepend(input).ASCII == input + digit + ASCII

    assert 'foo' + number + optional(whitespace) + word == number.append(whitespace.optional).prepend('foo').append(word)
    assert (
        optional(whitespace) + group(either(repeat('a'), 'b')) + if_followed_by(word) ==
        whitespace.optional.append(literal('a').repeat.or_('b').unnamed).if_followed_by(word) ==
        whitespace.optional + repeat('a').or_('b').unnamed + if_followed_by(word)
    )


def test_no_duplicate_flags():
    r = lineStart + word + '/' + '/' + lineEnd
    assert r.str() == r'(?m)^\w+//$'



from ..base.interface import *

del UNICODE

"Group: Premade"
email: EZRegex
"Matches an email"
version: EZRegex
"""The *official* regex for matching version numbers from https://semver.org/. It includes 5 groups that can be
matched/replaced: `major`, `minor`, `patch`, `prerelease`, and `buildmetadata`
"""
version_numbered: EZRegex
"Same as `version`, but it uses numbered groups for each version number instead of named groups"



# pyright: reportArgumentType = false
# pyright: reportUndefinedVariable = false
import warnings
from ..base import load_base, not_empty
from ..EZRegex import EZRegex
from .JavaScriptEZRegex import JavaScriptEZRegex
from string import digits

globals().update(load_base(JavaScriptEZRegex, lambda num_or_name, cur=...: fr'{cur}\k<{num_or_name}>'))

exactly = is_exactly = isExactly = JavaScriptEZRegex(lambda input, cur=...: r"^" + input + r"$")

def _group(input, name=None, cur=...):
    not_empty(input, 'group')
    if name is not None:
        not_empty(name, 'group', 'name')
    return f'{cur}({input})' if name is None else f'{cur}(?<{name}>{input})'

group = JavaScriptEZRegex(_group)

def _earlier_group(num_or_name, cur=...):
    not_empty(num_or_name, 'earlier_group', num_or_name)
    return fr'{cur}\{num_or_name}' if isinstance(num_or_name, int) or num_or_name in digits else fr'{cur}\k<{num_or_name}>'

earlier_group = earlierGroup = JavaScriptEZRegex(_earlier_group)

string_starts_with = string_start = stringStart = stringStartsWith = JavaScriptEZRegex(lambda input='', cur=...: r'^' + input + cur, string_anchor_used=True)
string_ends_with = string_end = stringEnd = stringEndsWith = JavaScriptEZRegex(lambda input='', cur=...: cur + input + r'$', string_anchor_used=True)

del ifExists
del if_exists
del ASCII

def _rgroup(num_or_name, cur=...):
    return fr'{cur}${num_or_name}' if isinstance(num_or_name, int) or num_or_name in digits else fr'{cur}$<{num_or_name}>'

# TODO: this probably doesn't all need to be here
def replace(string, rtn_str=True):
    class CustomFormatter(Formatter):
        def get_value(self, key, args, kwargs):
            return _rgroup(key, '')

    string = CustomFormatter().format(string)

    return string if rtn_str else JavaScriptEZRegex(string, sanatize=False, replacement=True)

rgroup = replace_group = JavaScriptEZRegex(_rgroup, replacement=True)
replace_entire = JavaScriptEZRegex('$&', replacement=True)



# pyright: reportArgumentType = false
# pyright: reportUndefinedVariable = false
from ..base import load_base, _parse_any_of_params, not_empty
from ..EZRegex import EZRegex
from .REZRegex import REZRegex

# As far as I can tell, this is the docs for this syntax:
# https://www.rdocumentation.org/packages/base/versions/3.6.2/topics/regex

globals().update(load_base(REZRegex, lambda num_or_name, cur=...: fr'{cur}\g{{{num_or_name}}}'))

# I can't figure out how flags work in R, so I'm just ignoring them
del line_starts_with
del lineStartsWith
del line_start
del lineStart
del line_ends_with
del lineEndsWith
del line_end
del lineEnd

del if_proceded_by
del ifProcededBy
del if_not_proceded_by
del ifNotProcededBy
del if_preceded_by
del ifPrecededBy
del if_not_preceded_by
del ifNotPrecededBy
del if_enclosed_with
del ifEnclosedWith

del ASCII
del DOTALL
del IGNORECASE
del LOCALE
del MULTILINE
del UNICODE

del exactly
del is_exactly
del isExactly

del earlier_group
del earlierGroup

punctuation = REZRegex(r'[' + REZRegex.escape(']`~!@#$%^&*()-_=+[{}\\|;:\'",<.>/?') + r']')

def _any_of(*inputs, chars=None, split=None, cur=...):
    chars, inputs = _parse_any_of_params(*inputs, chars=chars, split=split)

    if chars:
        if ']' in inputs:
            # You can't escape ] in an R character class, it has to be first
            inputs.remove(']')
            inputs.insert(0, ']')
        cur += r'['
        for i in inputs:
            cur += i
        cur += r']'
    else:
        cur += r'(?:'
        for i in inputs:
            cur += i
            cur += '|'
        cur = cur[:-1]
        cur += r')'
    return cur
any_of = anyOf = REZRegex(_any_of)

chunk = REZRegex(r'[^\n]+')

# Matches any single character except line break characters, like the dot, but is not affected by any options that make the dot match all characters including line breaks.
not_newline = REZRegex(r'\N')

def _group(input, name=None, cur=...):
    not_empty(input, 'group')
    if name is not None:
        raise ValueError('Named groups are not supported in R')
    return f'{cur}({input})'
group = REZRegex(_group)


def _rgroup(num, cur=...):
    return f'{cur}\\{num}'

# TODO: this probably doesn't all need to be here
def replace(string, rtn_str=True):
    class CustomFormatter(Formatter):
        def get_value(self, key, args, kwargs):
            return _rgroup(key, '')

    string = CustomFormatter().format(string)

    return string if rtn_str else REZRegex(string, sanatize=False, replacement=True)

rgroup = replace_group = REZRegex(_rgroup, replacement=True)
replace_entire = REZRegex('\\0', replacement=True)



from typing import TypedDict

from ezregex.EZRegex import EZRegex

Group = TypedDict(
    'Group',
    {
        'string': str,
        'end': int,
        'start': int,
        'color': str
    }
)

Match = TypedDict(
    'Match',
    {
        'string': str,
        'string HTML': str,
        'parts': list[list[str|None]],
        'end': int,
        'start': int,
        'color': str
    }
)

Matches = TypedDict(
    'Matches',
    {
        'match': Match,
        'unnamed groups': dict[int, Group],
        'named groups': dict[str, Group],
    }
)
APIStructure = TypedDict(
    "APIStructure",
    {
        'regex': str,
        'string': str,
        'string HTML': str,
        'parts': list[list[str|None]],
        'matches': list[Matches],
        'replaced': str | None,
        'split': list[str] | None,
    }
)

def api(
    pattern:EZRegex,
    replacement_pattern:EZRegex|str|None=None,
    test_string:str|None=None, *,
    replacement_count:int=0,
    split_count:int=0
) -> APIStructure: ...



import re
import sys
from os import replace
from typing import Any, Callable, Iterator

# from mypy_extensions import DefaultNamedArg, VarArg
from ezregex import EZRegex

from ..EZRegex import EZRegex

class PythonEZRegex(EZRegex):
    # Shadowing the re functions
    def compile(self, add_flags:bool=True): ...
    def search(self, string, pos:int=0, endpos:int=sys.maxsize) -> re.Match|None: ...
    def match(self, string, pos:int=0, endpos:int=sys.maxsize) -> re.Match|None: ...
    def fullmatch(self, string, pos:int=0, endpos:int=sys.maxsize) -> re.Match|None: ...
    def split(self, string, maxsplit:int=0) -> list: ...
    def findall(self, string, pos: int = 0, endpos: int = sys.maxsize) -> list: ...
    def finditer(self, string, pos: int = 0, endpos: int = sys.maxsize) -> Iterator[re.Match]: ...
    def sub(self, repl: Any | Callable[[re.Match], Any], string, count: int = 0): ...
    def subn(self, repl: Any | Callable[[re.Match], Any], string, count: int = 0): ...



"""
Generates a regex expression that matches everything in one list, but not the other
This file is a near-verbatim scrape of:
https://nbviewer.ipython.org/url/norvig.com/ipython/xkcd1313.ipynb
and
https://nbviewer.jupyter.org/url/norvig.com/ipython/xkcd1313-part2.ipynb
(if the links don't work, try them through the wayback machine)
Credit goes to Peter Norvig and Stefan Pochmann.
"""

import itertools
import random
import re
from collections import Counter, defaultdict

Set = frozenset # Data will be frozensets, so they can't be mutated.


cat = ''.join  # Join a sequence of strings with nothing between them

def OR(*regexes):
    """OR together component regexes. Ignore 'None' components.
    Allows both OR(a, b, c) and OR([a, b, c]), similar to max."""
    if len(regexes) == 1:
        regexes = regexes[0]
    return '|'.join(r for r in regexes if r)

def trivial(winners): return '^(' + OR(winners) + ')$'

def mistakes(regex, winners, losers):
    "The set of mistakes made by this regex in classifying winners and losers."
    return ({"Should have matched: " + W
             for W in winners if not re.search(regex, W)} |
            {"Should not have matched: " + L
             for L in losers if re.search(regex, L)})

def replacements(c):
    "All ways to replace character c with something interesting: for now, 'c' or '.'."
    return c if c in '^$' else c + '.'

def dotify(part):
    "Return all ways to replace a subset of chars in part with '.'."
    choices = map(replacements, part)
    return {cat(chars) for chars in itertools.product(*choices)}

def report(winners, losers):
    "Find a regex to match A but not B, and vice-versa.  Print summary."
    solution = findregex(winners, losers)
    assert not mistakes(solution, winners, losers)
    print('Chars: {}, ratio: {:.1f}, inputs: {}:{}'.format(
          len(solution), len(trivial(winners)) / len(solution) , len(winners), len(losers)))
    return solution

def subparts(word, max_len=5):
    "Return a set of subparts of word, consecutive characters up to length 5."
    return set(word[i:i+1+s] for i in range(len(word)) for s in range(max_len))

def matches(regex, strings):
    "Return a set of all the strings that are matched by regex."
    searcher = re.compile(regex).search
    return set(filter(searcher, strings))

def regex_parts(winners, losers):
    "Return parts that match at least one winner, but no loser."
    losers_str = '\n'.join(losers)
    def no_losers(part): return not re.compile(part, re.MULTILINE).search(losers_str)
    wholes = {'^' + w + '$' for w in winners}
    parts = {d for w in wholes for p in subparts(w) for d in dotify(p)}
    return wholes | set(filter(no_losers, parts))

def eliminate_dominated(covers):
    """Given a dict of {regex: {winner...}}, make a new dict with only the regexes
    that are not dominated by any others. A regex part p is dominated by p2 if p2 covers
    a superset of the matches covered by p, and rp is shorter."""
    newcovers = {}
    def signature(p): return (-len(covers[p]), len(p))
    for p in sorted(covers, key=signature):
        if not covers[p]: break # All remaining r must not cover anything
        # r goes in newcache if it is not dominated by any other regex
        if not any(covers[p2] >= covers[p] and len(p2) <= len(p)
                   for p2 in newcovers):
            newcovers[p] = covers[p]
    return newcovers

def select_necessary(covers):
    """Select winners covered by only one part; remove from covers.
    Return a pair of (covers, necessary)."""
    counts = Counter(w for p in covers for w in covers[p])
    necessary = {p for p in covers if any(counts[w] == 1 for w in covers[p])}
    if necessary:
        covered = {w for p in necessary for w in covers[p]}
        covers = {p: covers[p] - covered
                  for p in covers if p not in necessary}
        return covers, OR(necessary)
    else:
        return covers, None

def simplify_covers(covers, partial=None):
    "Eliminate dominated regexes, and select ones that uniquely cover a winner."
    previous = None
    while covers != previous:
        previous = covers
        covers = eliminate_dominated(covers)
        covers, necessary = select_necessary(covers)
        partial = OR(partial, necessary)
    return covers, partial

class BranchBound(object):
    "Hold state information for a branch and bound search."
    def __init__(self, winners, max_num_calls, k=4):
        self.cheapest = trivial(winners)
        self.calls    = max_num_calls
        self.k        = k

    def search(self, covers, partial=None):
        """Recursively extend partial regex until it matches all winners in covers.
        Try all reasonable combinations until we run out of calls."""
        if self.calls <= 0:
            return self.cheapest
        self.calls -= 1
        covers, partial = simplify_covers(covers, partial)
        if not covers: # Nothing left to cover; solution is complete
            self.cheapest = min(partial, self.cheapest, key=len)
        elif len(OR(partial, min(covers, key=len))) < len(self.cheapest):
            def score(p): return self.k * len(covers[p]) - len(p)
            best = max(covers, key=score) # Best part
            covered = covers[best] # Set of winners covered by best
            covers.pop(best)
            # Try with and without the greedy-best part
            self.search({c:covers[c]-covered for c in covers}, OR(partial, best))
            self.search(covers, partial)
        return self.cheapest

# TODO: Reimplement this eventually
# def consider_negative_lookahead(W, L):
#     "Return either SOLUTION[W, L] or negative lookup of SOLUTION[L, W], whichever is shorter."
#     solution = min(SOLUTION[W, L], '^(?!.*(' + SOLUTION[L, W] + '))',
#                    key=len)
#     assert not mistakes(solution, W, L)
#     return solution

class BranchBoundRandomRestart(BranchBound):
    def search(self, covers, partial=None):
        """Recursively extend partial regex until it matches all winners in covers.
        Try all reasonable combinations until we run out of calls."""
        if self.calls <= 0:
            return partial, covers
        self.calls -= 1
        covers, partial = simplify_covers(covers, partial)
        if not covers: # Nothing left to cover; solution is complete
            self.cheapest = min(partial, self.cheapest, key=len)
        elif len(OR(partial, min(covers, key=len))) < len(self.cheapest):
            # Try with and without the greedy-best component
            K = random.choice((2, 3, 4, 4, 4, 5, 6))
            F = random.choice((0.1, 0.1, 2.0))
            def score(c): return K * len(covers[c]) - len(c) + random.uniform(0., F)
            best = max(covers, key=score) # Best component
            covered = covers[best] # Set of winners covered by r
            covers.pop(best)
            self.search({c:covers[c]-covered for c in covers}, OR(partial, best))
            self.search(covers, partial)
        return self.cheapest

def repetitions(part):
    """Return a set of strings derived by inserting a single repetition character
    ('+' or '*' or '?'), after each non-special character.
    Avoid redundant repetition of dots."""
    splits = [(part[:i], part[i:]) for i in range(1, len(part)+1)]
    return {A + q + B
            for (A, B) in splits
            # Don't allow '^*' nor '$*' nor '..*' nor '.*.'
            if not (A[-1] in '^$')
            if not A.endswith('..')
            if not (A.endswith('.') and B.startswith('.'))
            for q in '*+?'}

def pairs(winners, special_chars=Set('*+?^$.[](){}|\\')):
    chars = Set(cat(winners)) - special_chars
    return {A+'.'+q+B
            for A in chars for B in chars for q in '*+?'}

def regex_covers(winners, losers, max_len=5):
    """Generate regex components and return a dict of {regex: {winner...}}.
    Each regex matches at least one winner and no loser."""
    losers_str = '\n'.join(losers)
    wholes = {'^'+winner+'$' for winner in winners}
    parts  = {d for w in wholes for p in subparts(w, max_len) for d in dotify(p)}
    reps   = {r for p in parts for r in repetitions(p)}
    pool   = wholes | parts | pairs(winners) | reps
    searchers = {p:re.compile(p, re.MULTILINE).search for p in pool}
    return {p: Set(filter(searchers[p], winners))
            for p in pool
            if not searchers[p](losers_str)}

def bb_findregex(winners, losers, calls=10000, restarts=10, max_len=5):
    "Find the shortest disjunction of regex components that covers winners but not losers."
    bb = BranchBoundRandomRestart(winners, calls)
    covers = eliminate_dominated(regex_covers(winners, losers, max_len))
    for _ in range(restarts):
        bb.calls = calls
        bb.search(covers.copy())
        if bb.calls > 0: # If search was not cut off, we have optimal solution
            return bb
    return bb

def findregex(winners:set, losers:set, calls:int=1000, restarts:int=3, max_len:int=5) -> str:
    """ The entrypoint into generating regex expressions from lists
        You give the function a set of things you want an expression to match (`winners`), and a set
        of things you want it not to match (`losers`), and it generates an expression that does so.

        `calls` sets the max number of iterations of the algorithm we do before we give up and return
        what we have
        `restarts` sets how many times we restart to algorithm in order to shake things up in case we've
        gone down a sub-optimal path
        `max_len` sets the maximum length of induvidual chunks that we OR together

        The algorithm runs for `calls` * `restarts` times, so tread lightly when setting those values.
        Increasing them causes the algorithm to run longer and find a shorter regex that meets the
        specifications, but it will also be significantly slower.

        Wrap this call in `raw()` to use it as an EZRegex expression

        Credit for this algorithm goes to Peter Norvig and Stefan Pochmann. See:
        https://nbviewer.jupyter.org/url/norvig.com/ipython/xkcd1313-part2.ipynb
    """
    return bb_findregex(set(winners), set(losers), calls=calls, restarts=restarts).cheapest
generate_regex = findregex



import colorsys
import re

# TODO:
# EZREgex todo:
# if overlapping groups, expand them all different amounts
# add classes to the spans so I can style them
# return the compiled replacement regex as well
# add code and string displays
# change the default background text color
# change the first color being white (give it a class?)
# remove empty spans

# These functions comprise the color algorithm
def toHtml(r, g, b):
    return f'#{r:02x}{g:02x}{b:02x}'

def toRgb(html: str) -> tuple:
    hex_color = html.lstrip('#')
    rgb = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))
    return rgb

def generate_colors(amt, s:float=1, v:float=1, offset:int=0):
    """ Generate `amt` number of colors evenly spaced around the color wheel
        with a given saturation and value
    """
    amt += 1
    return [toHtml(*map(lambda c: round(c*255), colorsys.hsv_to_rgb(*((offset + ((1/amt) * (i + 1))) % 1.001, s, v)))) for i in range(amt-1)]

def furthest_colors(html, amt:int=5, v_bias:float=0, s_bias:float=0):
    """ Gets the `amt` number of colors evenly spaced around the color wheel from the given color
        `v_bias` and `s_bias` are between 0-1 and offset the colors
    """
    amt += 1
    h, s, v = colorsys.rgb_to_hsv(*map(lambda c: c/255, toRgb(html)))

    return [toHtml(*map(lambda c: round(c*255), colorsys.hsv_to_rgb(*((h + ((1/amt) * (i + 1))) % 1.001, (s+s_bias) % 1.001, (v+v_bias) % 1.001)))) for i in range(amt-1)]


def api(pattern, replacement_pattern=None, test_string=None, *,
        replacement_count=0,
        split_count=0,
        # Can accept any valid CSS color
        default_text_color='black',
        container_tag='span',
        container_class='ezregex-container',
        match_class='ezregex-match',
        group_class='ezregex-group',
        unmatched_class='ezregex-unmatched',
        foreground_saturation = .75,
        foreground_value = 1,
        background_value_bias = .5,
        background_saturation_bias = .9,
    ):
    """ This functions like an API, even though it's not ever used as an actual API. It's used by
        the EZRegex frontend, as it loads this library locally. It made sense to put it in the
        library itself, becasue it could be useful for other purposes.
    """
    if isinstance(pattern, str):
        raise ValueError(f'The api `pattern` parameter must be of type EZRegex, recieved {type(pattern)}')

    # Get an inverse, if nessicary
    if test_string is None:
        test_string = pattern.inverse()
    matches = list(re.finditer(pattern._compile(), test_string))
    # found = bool(len(matches))

    json = {
        'regex': pattern._compile(),
        'string': test_string,
        'string HTML': ...,
        'parts': [],
        'matches': []
    }

    html_string = f'<{container_tag} class="{container_class}"><span style="color: {default_text_color};" class="{unmatched_class}">'
    parts = []
    global_cursor = 0
    all_matches = [m.span() for m in matches]
    # Map match spans to unique colors
    _colors = generate_colors(len(all_matches), s=foreground_saturation, v=foreground_value)
    match_colors = dict(zip(all_matches, _colors))

    for match in matches:
        all_groups = {match.span(i+1) for i in range(len(match.groups()))}
        named_groups = {i: match.span(i) for i in match.groupdict().keys()}
        # TODO: have named groups show their name and number instead of just their name
        # named_groups = {
        #     (cnt+1, ): match.span(cnt+1)
        #     for cnt, i in enumerate(match.groups())
        #     if i in match.groupdict().values()
        # }
        unnamed_groups = {
            cnt+1: match.span(cnt+1)
            for cnt, i in enumerate(match.groups())
            if i not in match.groupdict().values()
        }
        # Map group spans to unique colors
        # This gets equally spaced colors from the given color, so they're differentiable
        # and readable on a dark background
        colors = dict(zip(all_groups, furthest_colors(
            match_colors[match.span()],
            amt=len(all_groups),
            v_bias=background_value_bias,
            s_bias=background_saturation_bias
        )))
        cursor = match.span()[0]

        # First, get up until the match
        html_string += f'{test_string[global_cursor:cursor]}</span>'
        parts.append([None, None, test_string[global_cursor:cursor]])
        match_html = ''
        match_parts = []
        for g in sorted(all_groups, key=lambda x: x[0]):
            # This fixes the bug where overlapping groups get put in twice. By simply preventing
            # the cursor from moving backwards, we eliminate the latter (parent) group from being shown.
            if g[0] < cursor:
                continue

            # Print the match up until the group
            match_html += f'<span style="color: {match_colors[match.span()]};" class="{match_class}">{test_string[cursor:g[0]]}</span>'
            match_parts.append([match_colors[match.span()], None, test_string[cursor:g[0]]])

            # Print the group
            match_html += f'<span style="background-color: {colors[g]}; color: {match_colors[match.span()]};" class="{group_class}">{test_string[g[0]:g[1]]}</span>'
            match_parts.append([match_colors[match.span()], colors[g], test_string[g[0]:g[1]]])
            cursor = g[1]
        match_html += f'<span style="color: {match_colors[match.span()]};" class="{match_class}">{test_string[cursor:match.span()[1]]}</span>'
        match_parts.append([match_colors[match.span()], None, test_string[cursor:match.span()[1]]])
        global_cursor = match.span()[1]
        # Don't print after the group, cause there might be another match that covers it
        html_string += match_html
        parts += match_parts
        # to_slice = lambda t: f'({t[0]}:{t[1]})'
        match_json = {
            'match': {
                'string': match.group(),
                'string HTML': match_html,
                'parts': match_parts,
                'end': match.end(),
                'start': match.start(),
                "color": match_colors[match.span()],
            },
            "unnamed groups":{},
            "named groups":{},
        }

        for num, span in unnamed_groups.items():
            match_json['unnamed groups'][num] = {
                'string': match.group(num) or '',
                'end': span[1],
                'start': span[0],
                "color": colors[span],
            }

        for name, span in named_groups.items():
            match_json['named groups'][name] = {
                'string': match.group(name) or '',
                'end': span[1],
                'start': span[0],
                "color": colors[span],
            }
        json['matches'].append(match_json)

    # Don't forget to add any bit at the end that's not part of a match
    html_string += test_string[global_cursor:]
    parts.append([None, None, test_string[global_cursor:]])
    html_string += f'</span></{container_tag}>'

    # Remove any empty spans
    html_string = re.sub(r'<span[^>]*></span>', '', html_string)

    json['string HTML'] = html_string
    json['parts'] = parts
    if replacement_pattern is not None:
        json['replaced'] = re.sub(pattern.str(), replacement_pattern.str(), test_string, replacement_count)
    else:
        json['replaced'] = None
    json['split'] = re.split(pattern.str(), test_string, split_count)
    return json



import re

from ezregex.generate import *
from tests.data.groups import *
from tests.data.groups import _losers, _winners


def words(text):
    "All space-separated words in text."
    return Set(text.split())

def phrases(text, sep='/'):
    "All sep-separated phrases in text, uppercased and stripped."
    return Set(p.upper().strip() for p in text.split(sep))

# Tests
def test_new_parts():
    assert repetitions('a') == {'a+', 'a*', 'a?'}
    assert repetitions('ab') == {'a+b', 'a*b', 'a?b',
                                 'ab+', 'ab*', 'ab?'}
    assert repetitions('a.c') == {'a+.c', 'a*.c', 'a?.c',
                                  'a.c+', 'a.*c', 'a.?c',
                                  'a.+c', 'a.c*', 'a.c?'}
    assert repetitions('^a..d$') == {'^a+..d$', '^a*..d$', '^a?..d$',
                                     '^a..d+$', '^a..d*$', '^a..d?$'}
    assert pairs({'ab', 'c'}) == {
        'a.*a', 'a.*b', 'a.*c',
        'a.+a', 'a.+b', 'a.+c',
        'a.?a', 'a.?b', 'a.?c',
        'b.*a', 'b.*b', 'b.*c',
        'b.+a', 'b.+b', 'b.+c',
        'b.?a', 'b.?b', 'b.?c',
        'c.*a', 'c.*b', 'c.*c',
        'c.+a', 'c.+b', 'c.+c',
        'c.?a', 'c.?b','c.?c'}
    assert len(pairs({'1...2...3', '($2.34)', '42', '56', '7-11'})) == 8 * 8 * 3
    covers = regex_covers({'one', 'on'}, {'won', 'wuan', 'juan'})
    assert (eliminate_dominated(covers) == {'e': {'one'}, '^o': {'on', 'one'}})
    return 'test_new_parts passes'


def test_bb():
    assert OR(['a', 'b', 'c']) == OR('a', 'b', 'c') == 'a|b|c'
    assert OR(['a|b', 'c|d']) == OR('a|b', 'c|d') == 'a|b|c|d'
    assert OR(None, 'c') == 'c'
    covers1 = {'a': {'ann', 'abe'}, 'ab': {'abe'}}
    assert eliminate_dominated(covers1) == {'a': {'ann', 'abe'}}
    assert simplify_covers(covers1) == ({}, 'a')
    covers2 = {'a': {'abe', 'cab'}, 'b': {'abe', 'cab', 'bee'},
               'c': {'cab', 'cee'}, 'c.': {'cab', 'cee'}, 'abe': {'abe'},
               'ab': {'abe', 'cab'}, '.*b': {'abe', 'cab', 'bee'},
               'e': {'abe', 'bee', 'cee'}, 'f': {}, 'g': {}}
    assert eliminate_dominated(covers2) == simplify_covers(covers2)[0] == {
        'c': {'cab', 'cee'}, 'b': {'cab', 'abe', 'bee'}, 'e': {'cee', 'abe', 'bee'}}
    covers3 = {'1': {'w1'}, '.1': {'w1'}, '2': {'w2'}}
    assert eliminate_dominated(covers3) == {'1': {'w1'}, '2': {'w2'}}
    assert simplify_covers(covers3) in (({}, '2|1'), ({}, '1|2'))
    covers, nec = select_necessary({'a': {'abe'}, 'c': {'cee'}})
    assert covers == {} and (nec == 'c|a' or nec == 'a|c')
    assert {0, 1, 2} >= {1, 2}
    assert {1, 2} >= {1, 2}
    assert not ({1, 2, 4} >= {1, 3})
    return 'test_bb passes'

def test_generate():
    assert subparts('^it$') == {'^', 'i', 't', '$', '^i', 'it', 't$', '^it', 'it$', '^it$'}
    assert subparts('this') == {'t', 'h', 'i', 's', 'th', 'hi', 'is', 'thi', 'his', 'this'}
    subparts('banana') == {'a', 'an', 'ana', 'anan', 'b', 'ba', 'ban', 'bana',
                           'n', 'na', 'nan', 'nana'}

    assert dotify('it') == {'it', 'i.', '.t', '..'}
    assert dotify('^it$') == {'^it$', '^i.$', '^.t$', '^..$'}
    assert dotify('this') == {'this', 'thi.', 'th.s', 'th..', 't.is', 't.i.', 't..s', 't...',
                              '.his', '.hi.', '.h.s', '.h..', '..is', '..i.', '...s', '....'}
    # assert regex_parts({'win'}, {'losers', 'bin', 'won'}) == {
    #     '^win$', '^win', '^wi.', 'wi.',  'wi', '^wi', 'win$', 'win', 'wi.$'}
    assert regex_parts({'win'}, {'bin', 'won', 'wine', 'wit'}) == {'^win$', 'win$'}
    regex_parts({'boy', 'coy'},
                {'ahoy', 'toy', 'book', 'cook', 'boycott', 'cowboy', 'cod', 'buy', 'oy',
                 'foil', 'coyote'}) == {'^boy$', '^coy$', 'c.y$', 'coy$'}

    assert matches('a|b|c', {'a', 'b', 'c', 'd', 'e'}) == {'a', 'b', 'c'}
    assert matches('a|b|c', {'any', 'bee', 'succeed', 'dee', 'eee!'}) == {
        'any', 'bee', 'succeed'}

    assert OR(['a', 'b', 'c']) == 'a|b|c'
    assert OR(['a']) == 'a'

    assert words('this is a test this is') == {'this', 'is', 'a', 'test'}

    assert findregex({"ahahah", "ciao"},  {"ahaha", "bye"}) == 'a.$'
    # This is *an* option, not the only option
    # assert findregex({"this", "that", "the other"}, {"one", "two", "here", "there"}) == 'h..$'
    assert findregex({'boy', 'coy', 'toy', 'joy'}, {'ahoy', 'buy', 'oy', 'foil'}) == '^.oy'

    assert not mistakes('a|b|c', {'ahoy', 'boy', 'coy'}, {'joy', 'toy'})
    assert not mistakes('^a|^b|^c', {'ahoy', 'boy', 'coy'}, {'joy', 'toy', 'kickback'})
    assert mistakes('^.oy', {'ahoy', 'boy', 'coy'}, {'joy', 'ploy'}) == {
        "Should have matched: ahoy",
        "Should not have matched: joy"}
    return 'tests pass'



def test_generate_auto():
    for w, l in zip(_winners, _losers):
        r = re.compile(generate_regex(w, l, 500, restarts=2))
        for i in w:
            assert r.search(i), f"`{r}` is not in `{i}`"
        for i in l:
            assert not r.search(i), f"`{r}` is in `{i}`"
        # print(r, 'is good')



from string import Formatter
from types import EllipsisType
from typing import Callable

from .elements import base, psuedonymns, not_empty, input_not_empty, _parse_any_of_params


def load_base(cls, rgroup_func: Callable[[int|str, EllipsisType], str], replace_entire_func=...) -> dict:
    rtn = {}
    for name, kwargs in base.items():
        rtn[name] = cls(**kwargs)

    # I'm creating this here, so we don't have to reimplement both of them every time
    def replace(string, rtn_str=True):
        class CustomFormatter(Formatter):
            def get_value(self, key, args, kwargs):
                return rgroup_func(key, '')

        string = CustomFormatter().format(string)

        return string if rtn_str else cls(string, sanatize=False, replacement=True)


    rtn['replace'] = replace
    rtn['rgroup'] = cls(rgroup_func, replacement=True)
    rtn['replace_entire'] = cls(
        lambda cur=...: rgroup_func(0, cur=cur) if replace_entire_func is Ellipsis else replace_entire_func,
        replacement=True
    )

    for original, aliases in psuedonymns.items():
        for alias in aliases:
            rtn[alias] = rtn[original]

    return rtn



import re
from copy import copy
from dataclasses import dataclass
from random import choice, choices, randint
from warnings import warn


def unsanitize(string):
    return re.sub(r'\\([' + re.escape(')([]{}+*$^-\\?| ,') + r'])', r'\g<1>', string)

def randbool():
    return bool(randint(0, 1))

def _randWord(randomlyGenerate=False):
    if randomlyGenerate:
        return ''.join(choices(_letters + '_', k=randint(1, _alot)))
    else:
        return 'word'

# Inverting helpers
_alot         = 6
_digits       = '0123456789'
_letters      = 'abcdefghijklmnopqrstuvwxyz'
_letters     += _letters.upper()
_punctuation  = "./;=-&%$#@~"
_whitespace   = '  '
_everything   = _digits + _letters + _punctuation + _whitespace + '_'

# notParens = matchMax(anyCharExcept('(', ')'))  # + ifNotPreceededBy(r'\\'))
# inner = atLeastNone(either('(' + optional('?:') + notParens + ')', notParens))
# prevThing = namedGroup('prev', anyof('(?:' + inner + ')', '(' + inner + ')', ifNotPreceededBy(r'\\') + anything))
prevThing = r'(?P<prev>(?:\(\?\:(?:(?:\((?:\?\:)?(?:[^\(\)])+\)|(?:[^\(\)])+))*\)|\((?:(?:\((?:\?\:)?(?:[^\(\)])+\)|(?:[^\(\)])+))*\)|(?<!\\).))'

tillCloseParen = r'(?P<stuff>.+)\)'

@dataclass
class _invertRegexes:
    flags        = re.compile(r"\A\(\?[aiLmsux]\)")
    word         = re.compile(r'\\w\+')
    s            = re.compile(r'\\s')
    digit        = re.compile(r'\\d')
    char         = re.compile(r'\\w')
    any          = re.compile(r'(?<!\\)\.')
    newline      = re.compile(r'\\n')
    carriage     = re.compile(r'\\r')
    tab          = re.compile(r'\\t')
    vert         = re.compile(r'\\v')
    f            = re.compile(r'\\f')
    stringStart  = re.compile(r'\\A')
    lineStart    = re.compile(r'\^')
    stringEnd    = re.compile(r'\\Z')
    lineEnd      = re.compile(r'\$')
    matchExactly = re.compile(r'\^(.+)\$')  # \^stuff\$
    # '[' + atLeastOne(namedGroup('start', char) + '-' + namedGroup('end', char)) + ']'
    anyBetween   = re.compile(r'\[(?:(?P<start>.)\-(?P<end>.))+\]')
    # '(' + namedGroup('stuff', chunk()) + '){' + namedGroup('amt', number()) + '}'
    matchExactAmt = re.compile(prevThing + r'\{(?P<amt>\d+)\}')  # (stuff){3}
    # '(' + namedGroup('stuff', chunk()) + '){' + namedGroup('min', number()) + ',' + optional(space()) + '}'
    matchMore = re.compile(prevThing + r'\{(?P<min>\d+),( )?\}')  # (stuff){3,}
    # optional  = re.compile(prevThing + r'\?(?![\:P\<\!])')  # (stuff)?
    optionalGreedy  = re.compile(prevThing + r'\?(?![\:P\<\!])')  # (stuff)?
    optionalNonGreedy  = re.compile(prevThing + r'\?\?(?![\:P\<\!])')  # (stuff)??
    atLeastNoneGreedy       = re.compile(prevThing + r'\*')  # (stuff)*
    atLeastNoneNonGreedy       = re.compile(prevThing + r'\*\?')  # (stuff)*?
    atLeastOneGreedy     = re.compile(fr'{prevThing}\+')  # (stuff)+
    atLeastOneNonGreedy     = re.compile(fr'{prevThing}\+\?')  # (stuff)+?
    # '(' + namedGroup('stuff', chunk()) + '){' + namedGroup('min', number()) + ',' + optional(space()) + namedGroup('max', number()) + '}'
    matchRangeGreedy = re.compile(prevThing + r'\{(?P<min>\d+),( )?(?P<max>\d+)\}')  # (stuff){3,4}
    matchRangeNonGreedy = re.compile(prevThing + r'\{(?P<min>\d+),( )?(?P<max>\d+)\}\?')  # (stuff){3,4}?
    # either('(?:', '(') + group(atLeast1(notParens, greedy=False)) + '|' + group(atLeast1(char, greedy=False)) + ')'
    either    = re.compile(r'(?:\(\?\:|\()((?:(?:[^\(\)](?<!\\))+)+?)\|(.+?)\)')  # (stuff|things)
    notWhitespace = re.compile(r'\\S')
    notDigit      = re.compile(r'\\D')
    notWord       = re.compile(r'\\W')
    # '[' + ifNotFollowedBy('^') + group(atLeastOne(anyCharExcept(','))) + ']'
    anyChars    = re.compile(r'\[(?!\^)((?:[^\,])+)\]')  # [st]
    anyChars2    = re.compile(r'\[(?!\^)((?:.(?:\,)?)+)\]')  # [s,t]
    # '[' + ifFollowedBy('^') + group(atLeastOne(char)) + ']'
    notAnyChars = re.compile(r'\[(?=\^)(.+)\]')  # [^s,t]
    # group(either(ifPrecededBy('(?:'), ifPrecededBy('(')) + matchMax(anyCharExcept(r'|\)')) + '|' + matchMax(anyCharExcept(r'|\)')) + optional('|') + ifFollowedBy(')'))
    anyOf = re.compile(r'((?:(?<=\(\?\:)|(?<=\())(?:.\|)+.(?=\)))')  # (stuff|things|otherThings)
    ifThing = re.compile(r'\(\?\=' + tillCloseParen)  # (?=stuff)
    ifNotThing       = re.compile(r'\(\?!' + tillCloseParen)  # (?!stuff)
    ifPrecededBy     = re.compile(r'\(\?\<\=' + tillCloseParen)
    ifNotPrecededBy  = re.compile(r'\(\?\<\!' + tillCloseParen)
    ifNotPrecededBy2 = re.compile(r'\(\?\!\=' + tillCloseParen)
    # '(?:' + group(matchMax(notAnyChars(')', '(', '?'))) + ')'
    passiveGroup     = re.compile(r'\(\?\:((?:[^\(\?])+)\)')  # (?:stuff)
    group = re.compile(r'\((?!\?)([^\)\(\?]+)\)')  # (stuff)
    namedGroup = re.compile(r'\(\?P\<(?P<name>\w+)\>(?P<stuff>.+)\)')  # (P?<name>stuff)


def invertRegex(regex:str, tries=10) -> str:
    """ 'Inverts' a regex expression.
        Gives an example of something that would match the given regex
        If tries is positive, it's guarenteed to return an expression that
        correctly matches the given expression, or raise an Error
    """
    # warn('This backend only works most of the time. The default backend is prefered.')
    # So we can compare later and make sure it works
    original = copy(regex)

    # So you can pass in an EasyRegex chain
    regex = str(regex)

    # Because I hate everyone and this is a stupid bug and it makes NO sense
    regex = re.sub(r'\\(?![AZbBcsSdDwWx0\|QEnrtvfox])', r'\\\\', regex)

    #* The order of these matter

    # Very first thing, remove any flags at the beginning
    # TODO: keep track of what flags are asserted. It's probably important?
    regex = _invertRegexes.flags.sub('', regex)

    regex = _invertRegexes.stringStart.sub('', regex)  # TODO
    regex = _invertRegexes.stringEnd.sub('', regex)  # TODO
    regex = _invertRegexes.lineEnd.sub('', regex)  # TODO

    # Single Characters
    regex = _invertRegexes.word.sub(_randWord(), regex)
    regex = _invertRegexes.s.sub(_whitespace, regex)
    regex = _invertRegexes.digit.sub(choice(_digits), regex)
    regex = _invertRegexes.char.sub(choice(_letters + '_'), regex)
    regex = _invertRegexes.any.sub(choice(_everything), regex)

    # Explicit Characters
    regex = _invertRegexes.newline.sub('\n', regex)
    regex = _invertRegexes.carriage.sub('\r', regex)
    regex = _invertRegexes.tab.sub('\t', regex)
    regex = _invertRegexes.vert.sub('\v', regex)
    regex = _invertRegexes.f.sub('\f', regex)

    # Matching
    regex = _invertRegexes.matchExactly.sub(r'\1', regex)

    # Amounts
    regex = _invertRegexes.matchRangeNonGreedy.sub(lambda m: m.group('prev') * randint(int(m.group('min')), int(m.group('max'))), regex)
    regex = _invertRegexes.matchRangeGreedy.sub(lambda m: m.group('prev') * randint(int(m.group('min')), int(m.group('max'))), regex)

    regex = _invertRegexes.atLeastOneNonGreedy.sub(r"\g<prev>" * randint(1, _alot), regex)
    regex = _invertRegexes.atLeastOneGreedy.sub(r"\g<prev>" * randint(1, _alot), regex)
    regex = _invertRegexes.matchExactAmt.sub(lambda m: m.group('prev') * randint(int(m.group('amt')), int(m.group('amt'))), regex)
    regex = _invertRegexes.matchMore.sub(lambda m: m.group('prev') * randint(int(m.group('min')), int(m.group('min')) + _alot), regex)

    # Not Chuncks
    regex = _invertRegexes.notWhitespace.sub(choice(_digits  + _letters + _punctuation + '_'), regex)
    regex = _invertRegexes.notDigit.sub(choice(_letters + _whitespace  + _punctuation + '_'), regex)
    regex = _invertRegexes.notWord.sub(choice(_punctuation + _whitespace), regex)

    # Conditionals
    regex = _invertRegexes.ifThing.sub(r'\g<stuff>', regex)
    regex = _invertRegexes.ifNotThing.sub(r'\g<stuff>', regex)
    regex = _invertRegexes.ifPrecededBy.sub(r'\g<stuff>', regex)
    regex = _invertRegexes.ifNotPrecededBy.sub(r'\g<stuff>', regex)
    regex = _invertRegexes.ifNotPrecededBy2.sub(r'\g<stuff>', regex)

    # Optionals
    regex = _invertRegexes.atLeastNoneNonGreedy.sub(r'\g<prev>' * randint(0, _alot), regex)
    regex = _invertRegexes.atLeastNoneGreedy.sub(r'\g<prev>' * randint(0, _alot), regex)
    regex = _invertRegexes.either.sub(r'\g<1>' if randint(0,1) else r'\g<2>', regex)
    # TODO: This isn't wrong, but it also isn't right. It only uses the start or the end, instead of the range
    regex = _invertRegexes.anyBetween.sub(r'\g<start>' if randint(0,1) else r'\g<end>', regex)
    regex = _invertRegexes.notAnyChars.sub(lambda m: choice(list(set(_everything).difference(list(m.group(1))))), regex)
    regex = _invertRegexes.anyChars.sub(lambda m: choice(list(m.group(1))), regex)
    regex = _invertRegexes.anyChars2.sub(lambda m: choice(m.groups()).replace(',', ''), regex)

    regex = _invertRegexes.lineStart.sub('', regex)  # TODO

    def tmp(m):
        options = m.groups()[0].split('|')
        return choice(options)
    regex = _invertRegexes.anyOf.sub(tmp, regex)

    # Groups
    def denest(regex, replace, current):
        new = regex.sub(replace, current)
        if current != new:
            new = denest(regex, replace, new)
        return new

    regex = denest(_invertRegexes.optionalNonGreedy, r'\g<prev>' if randint(0,1) else '', regex)
    regex = denest(_invertRegexes.optionalGreedy, r'\g<prev>' if randint(0,1) else '', regex)
    regex = denest(_invertRegexes.namedGroup, r'\g<stuff>', regex)
    regex = denest(_invertRegexes.passiveGroup, r'\1', regex)
    regex = denest(_invertRegexes.group, r'\1', regex)

    # referenceGroup = EasyRegexSingleton(lambda cur, name:        f'{cur}(?P={name})') # TODO))

    # Undo what we did earlier
    regex = re.sub(r'\\\\', r'\\', regex)
    regex = unsanitize(regex)

    # If the regex passed in isn't in what we generated, try again
    if re.search(str(original), regex) is not None or tries < 0:
        return regex
    else:
        if not tries:
            raise NotImplementedError(f"Can't invert the given regex. You can try again, which may work, or submit a bug report by emailing smartycope@gmail.com and including your pattern")
        else:
            return invertRegex(original, tries=tries-1)



import re

import pytest

from ezregex import *

strictness=20
dontIncludePassed=True
invertBackend='re_parser'
invert_tries=1
offset = 2


def test_invert():
    import ezregex as er
    regex = (er.digit + er.word)
    assert re.search(regex.str(), ~regex)

# TODO: Finish this
# assert digit * ... == matchMax(digit)

# TODO: Debug this
# assert digit | word == anyof(digit, word), f"{digit | word} != {anyof(digit, word)}"
# assert 7 | digit == anyof(7, digit), f"{7 | digit} != {anyof(digit, 7)}"
# assert re.search(str(digit | word | '1'), '1') == re.search(anyof(digit, word, 1).str(), '1')

def test_unary_plus():
    assert +digit == matchMax(digit)

def test_concat_operators():
    assert anything + word == anything << word
    assert anything + word == anything >> word

def test_square_brackets():
    assert digit[2, 3] == match_range(2, 3, digit)
    assert digit[2, ...] == digit[2,] == digit[2, None] == digit[2] == match_at_least(2, digit)
    assert digit[..., 2] == digit[0, 2] == digit[None, 2] == match_at_most(2, digit)
    assert digit[...] == digit[0, ...] == digit[None] == at_least_0(digit)
    assert digit[1, ...] == digit[1] == digit[1,] == digit[1, None] == at_least_1(digit)
    # expr[...:end_expr] is equivalent to ZeroOrMore(expr, stop_on=end_expr)
    # assert digit[...:'foo'] == digit[None:'foo'] == digit[,'foo'] ==

def test_iadd():
    test = digit
    test += 6
    assert test == digit + 6, f"{test} != {digit + 6}"

def test_imul():
    test = digit
    test *= 3
    assert test == digit * 3

def test_mul():
    assert 2 * digit == digit * 2

def test_mod():
    assert digit % 'sldkj' is None, f'{digit % "sldkj"} != None'
    assert digit % '77sdsf88' == re.search('77sdsf88', digit.str())

# Test f-strings
# As of right now, these may work, but won't necissarily work
# assert f'|{word}|' == '|' + word + '|'
# assert f'|{word}|{number}|' == '|' + word + '|' + number + '|'
# assert f"|{word}{number + ifFollowedBy(word)}|" == '|' + word + number + ifFollowedBy(word) + '|'
# assert f'|{rgroup("g")}|{replace_group}|' == '|' + rgroup('g') + '|' + replace_entire + '|'

# Replacement strings are, in fact, a thing
# print(version_info)

# no idea why this doesnt work.
# assert (anything + word) * 3 == '.\w+' * 3, f"'{(anything + word) * 3}' != '{'.\w+'*3}'"




# From 1-100, 1 is easy, 100 is hard
# difficulty = 1
# runTests(
# # These should remain on, for the GitHub automated tests
# singletons=True,
# _invert=False,
# replacement=True,
# operators=True,
# _generate=True,
# # These display for you to check that they look correct
# testMethod=False,
# _api=False,
# # Settings
# strictness=difficulty,
# invert_tries=101-difficulty,
# dontIncludePassed=True,
# invertBackend='re_parser',
# )



# pyright: reportOperatorIssue = false
from re import escape
from string import digits
from sys import version_info


def input_not_empty(func, name, not_empty=0):
    def rtn(*args, **kwargs):
        if not len(str(args[not_empty])):
            raise ValueError(f'Input parameter of {name} cannot be empty')
        return func(*args, **kwargs)
    return rtn

def not_empty(param, func, param_name='input'):
    if not len(str(param)):
        raise ValueError(f'Parameter {param_name} in {func} cannot be empty')


def match_range(min, max, input, greedy=True, possessive=False, cur=...):
    if ((not greedy) and possessive):
        raise ValueError('match_range can\'t be non-greedy AND possessive at the same time')
    not_empty(input, 'match_range')

    s = cur
    if len(input):
        s += r'(?:' + input + r')'
    s += r'{' + str(min) + r',' + str(max) + r'}'
    if not greedy:
        s += r'?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += r'+'
    return s

def at_least_one(input, greedy=True, possessive=False, cur=...):
    if ((not greedy) and possessive):
        raise ValueError('at_least_one can\'t be non-greedy AND possessive at the same time')
    not_empty(input, 'at_least_one')

    s = cur
    if len(input) > 1:
        s += fr'(?:{input})+'
    else:
        if len(input) == 1:
            s += fr'{input}+'
    if not greedy:
        s += '?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += '+'
    return s

def at_least_none(input, greedy=True, possessive=False, cur=...):
    if ((not greedy) and possessive):
        raise ValueError('at_least_none can\'t be non-greedy AND possessive at the same time')
    not_empty(input, 'at_least_none')

    s = cur
    if len(input) > 1:
        s += fr'(?:{input})*'
    else:
        if len(input) == 1:
            s += fr'{input}*'
    if not greedy:
        s += '?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += '+'
    return s

def optional(input, greedy=True, possessive=False, cur=...):
    if ((not greedy) and possessive):
        raise ValueError('optional can\'t be non-greedy AND possessive at the same time')
    not_empty(input, 'optional')

    s = cur
    if len(input) > 1:
        s += fr'(?:{input})?'
    else:
        if len(input) == 1:
            s += fr'{input}?'
    if not greedy:
        s += r'?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += r'+'
    return s

def _parse_any_of_params(*inputs, chars=None, split=None):
    if split and len(inputs) != 1:
        raise ValueError("Please don't specifiy split and pass multiple inputs to anyof")
    elif split:
        inputs = list(inputs[0])
    elif len(inputs) == 1 and split is None and chars is not False:  # None means auto
        chars = True
        inputs = list(inputs[0])
    elif len(inputs) == 1 and split is None:
        inputs = list(inputs[0])
    elif len(inputs) > 1 and chars is None and all(map(lambda s: len(str(s)) == 1, inputs)):
        chars = True

    return chars, inputs

def any_of(*inputs, chars=None, split=None, cur=...):
    chars, inputs = _parse_any_of_params(*inputs, chars=chars, split=split)

    if chars:
        cur += r'['
        for i in inputs:
            cur += i
        cur += r']'
    else:
        cur += r'(?:'
        for i in inputs:
            cur += i
            cur += '|'
        cur = cur[:-1]
        cur += r')'
    return cur

def any_char_except(*inputs, cur=...):
    # If it's just a string, split it up
    if len(inputs) == 1 and len(inputs[0]) > 1:
        inputs = list(inputs[0])

    cur += r'[^'
    for i in inputs:
        cur += i
    cur += r']'
    return cur

def each(*inputs, cur=...):
    inputs = list(inputs)
    last = inputs.pop()
    s = cur
    for i in inputs:
        s += fr'(?={i})'
    s += last
    return s

def any_between(char, and_char, cur=...):
    not_empty(char, 'any_between', 'char')
    not_empty(and_char, 'any_between', 'and_char')
    return cur + r'[' + char + r'-' + and_char + r']'

def match_max(input, cur=...):
    not_empty(input, 'match_max')
    return cur + r'(?:' + input + r')' + r'+'

def match_num(num, input, cur=...):
    not_empty(input, 'match_num')
    return cur + r'(?:' + input + r')' + r'{' + str(num) + r'}'

def match_more_than(min, input, cur=...):
    not_empty(input, 'match_more_than')
    return cur + r'(?:' + input + r')' + r'{' + str(int(min) + 1) + r',}'

def match_at_least(min, input, cur=...):
    not_empty(input, 'match_at_least')
    return cur + r'(?:' + input + r')' + r'{' + str(min) + r',}'

def match_at_most(max, input, cur=...):
    not_empty(input, 'match_at_most')
    return cur + r'(?:' + input + r')' + r'{0,' + str(max) + r'}'

def either(input, or_input, cur=...):
    not_empty(input, 'either')
    return cur + rf'(?:{input}|{or_input})'

def any_except(input, type='.*', cur=...):
    not_empty(input, 'any_except')
    return cur + f'(?!{input}){type}'

def if_proceded_by(input, cur=...):
    not_empty(input, 'if_proceded_by')
    return fr'{cur}(?={input})'

def if_not_proceded_by(input, cur=...):
    not_empty(input, 'if_not_proceded_by')
    return fr'{cur}(?!{input})'

def if_preceded_by(input, cur=...):
    not_empty(input, 'if_preceded_by')
    return fr'(?<={input}){cur}'

def if_not_preceded_by(input, cur=...):
    not_empty(input, 'if_not_preceded_by')
    return fr'(?<!{input}){cur}'

def if_enclosed_with(open, stuff, close=..., cur=...):
    not_empty(input, 'if_enclosed_with')
    if close is Ellipsis:
        close = open
    return fr'((?<={open}){stuff}(?={open if close is None else close}))'

def group(input, name=None, cur=...):
    not_empty(input, 'group')
    if name is not None:
        not_empty(name, 'group', 'name')
    return f'{cur}({input})' if name is None else f'{cur}(?P<{name}>{input})'

def passive_group(input, cur=...):
    not_empty(input, 'passive_group')
    return f'{cur}(?:{input})'

def earlier_group(num_or_name, cur=...):
    not_empty(num_or_name, 'earlier_group', num_or_name)
    return f'{cur}\\{num_or_name}' if isinstance(num_or_name, int) or num_or_name in digits else f'{cur}(?P={num_or_name})'

def if_exists(num_or_name, does, doesnt, cur=...):
    not_empty(num_or_name, 'earlier_group', num_or_name)
    return f'{cur}(?({num_or_name}){does}{"|" + str(doesnt) if doesnt is not None else ""})'


base = {
    # Positional
    # TODO: enforce these only being at the end or beginning of a chain -- maybe
    'string_starts_with': {'definition': lambda input='', cur=...: r'\A' + input + cur},
    'string_ends_with':   {'definition': lambda input='', cur=...: input + r'\Z' + cur},
    # Always use the multiline flag, so as to distinguish between start of a line vs start of the string
    'line_starts_with':   {'definition': lambda input='', cur=...: r'^' + input + cur, 'flags':'m'},
    'line_ends_with':     {'definition': lambda input='', cur=...: cur + input + r'$', 'flags':'m'},
    'word_boundary':      {'definition': r'\b'},
    'not_word_boundary':  {'definition': r'\B'},

    # Literals
    'tab':                {'definition': r'\t'},
    'space':              {'definition': r' '},
    'space_or_tab':       {'definition': r'[ \t]'},
    'new_line':           {'definition': r'\n'},
    'carriage_return':    {'definition': r'\r'},
    'quote':              {'definition': r'(?:\'|"|`)'},
    'vertical_tab':       {'definition': r'\v'},
    'form_feed':          {'definition': r'\f'},
    'comma':              {'definition': r'\,'},
    'period':             {'definition': r'\.'},
    'underscore':         {'definition': r'_'},
    'any_between':        {'definition': any_between},

    # Not Literals
    'not_whitespace':     {'definition': r'\S'},
    'not_digit':          {'definition': r'\D'},
    'not_word':           {'definition': r'\W'},

    # Catagories
    'whitespace':         {'definition': r'\s'},
    'whitechunk':         {'definition': r'\s+'},
    'digit':              {'definition': r'\d'},
    'number':             {'definition': r'\d+'},
    'word':               {'definition': r'\w+'},
    'word_char':          {'definition': r'\w'},
    'anything':           {'definition': r'.'},
    'chunk':              {'definition': r'.+'},
    'uppercase':          {'definition': r'[A-Z]'},
    'lowercase':          {'definition': r'[a-z]'},
    'letter':             {'definition': r'[A-Za-z]'},
    'hex_digit':          {'definition': r'[0-9a-fA-F]'},
    'oct_digit':          {'definition': r'[0-7]'},
    # TODO: is there a more formal definition of this or something?
    'punctuation':        {'definition': r'[' + escape('`~!@#$%^&*()-_=+[{]}\\|;:\'",<.>/?¢]') + r']'},
    'controller':         {'definition': r'[\x00-\x1F\x7F]'},
    'printable':          {'definition': r'[\x21-\x7E]'},
    'printable_and_space':{'definition': r'[\x20-\x7E]'},
    'alpha_num':          {'definition': r'[A-Za-z0-9_]'},
    'unicode':            {'definition': lambda name, cur=...: fr'\N{name}'},

    # Amounts
    'repeat':             {'definition': match_max},
    'amt':                {'definition': match_num},
    'more_than':          {'definition': match_more_than},
    'at_least':           {'definition': match_at_least},
    'at_most':            {'definition': match_at_most},
    'between':            {'definition': match_range},
    'at_least_one':       {'definition': at_least_one},
    'at_least_none':      {'definition': at_least_none},

    # Choices
    'optional':           {'definition': optional},
    'either':             {'definition': either},
    'any_of':             {'definition': any_of},
    'any_char_except':    {'definition': any_char_except},
    'any_except':         {'definition': any_except},
    'each':               {'definition': each},

    # Conditionals
    'if_proceded_by':     {'definition': if_proceded_by},
    'if_not_proceded_by': {'definition': if_not_proceded_by},
    'if_preceded_by':     {'definition': if_preceded_by},
    'if_not_preceded_by': {'definition': if_not_preceded_by},
    'if_enclosed_with':   {'definition': if_enclosed_with},

    # Grouping
    'group':              {'definition': group},
    'passive_group':      {'definition': passive_group},
    'earlier_group':      {'definition': earlier_group},
    'if_exists':          {'definition': if_exists},

    # Premade
    # TODO: a chunk of literally anything/chunk of literally anything except ...
    'literally_anything': {'definition': r'(?:.|\n)'},
    'signed':             {'definition': r'(?:(?:\-|\+))?\d+'},
    'unsigned':           {'definition': r'\d+'},
    'plain_float':        {'definition': r'(?:(?:\-|\+))?\d+\.(?:\d+)?'},
    'full_float':         {'definition': r'(?:(?:\-|\+))?\d+\.(?:\d+)?(?:e(?:(?:\-|\+))?\d+)?'},
    'int_or_float':       {'definition': r'(?:(?:\-|\+))?\d+\.(?:\d+)?(?:e(?:(?:\-|\+))?\d+)?(?:\-)?\d+(?:\.(?:\d+)?)?'},
    'ow':                 {'definition': r'\s*'},

    # Misc.
    'is_exactly':         {'definition': lambda input, cur=...: r"\A" + input + r'\Z'},
    'literal':            {'definition': lambda input, cur=...: cur + input},
    'raw':                {'definition': lambda regex, cur=...: str(regex), 'sanatize': False},

    # Flags
    'ASCII':              {'definition': lambda cur=...: cur, 'flags': 'a'},
    'DOTALL':             {'definition': lambda cur=...: cur, 'flags': 's'},
    'IGNORECASE':         {'definition': lambda cur=...: cur, 'flags': 'i'},
    'LOCALE':             {'definition': lambda cur=...: cur, 'flags': 'L'},
    'MULTILINE':          {'definition': lambda cur=...: cur, 'flags': 'm'},
    'UNICODE':            {'definition': lambda cur=...: cur, 'flags': 'u'},
}


psuedonymns = {
    'repeat': ('matchMax', 'match_max'),
    'at_most': ('matchAtMost', 'atMost', 'match_at_most',),
    'amt': ('matchNum', 'matchAmt', 'match_amt', 'match_num', 'num',),
    'between': ('matchRange', 'match_range'),
    'more_than': ('matchMoreThan', 'match_greater_than', 'matchGreaterThan', 'moreThan', 'match_more_than',),
    'at_least': ('matchAtLeast', 'match_min', 'matchMin', 'atLeast', 'match_at_least',),
    'line_starts_with': ('lineStartsWith', 'line_start', 'lineStart',),
    'string_starts_with': ('stringStartsWith', 'string_start', 'stringStart',),
    'line_ends_with': ('lineEndsWith', 'line_end', 'lineEnd',),
    'string_ends_with': ('stringEndsWith', 'string_end', 'stringEnd',),
    'chunk': ('stuff',),
    'whitechunk': ('whiteChunk', 'white'),
    'anything': ('anychar', 'anyChar', 'char'),
    'letter': ('alpha',),
    'alpha_num': ('alphanum'  , 'alpha_num',),
    'any_between': ('anyBetween',),
    'word_char': ('wordChar',),
    'hex_digit': ('hexDigit', 'hex',),
    'oct_digit': ('octDigit',),
    'new_line': ('newline', 'newLine',),
    'space_or_tab': ('spaceOrTab',),
    'carriage_return': ('carriageReturn',),
    'vertical_tab': ('verticalTab',),
    'form_feed': ('formFeed',),
    'period': ('dot',),
    'int_or_float': ('intOrFloat',),
    'not_whitespace': ('notWhitespace',),
    'not_digit': ('notDigit',),
    'not_word': ('notWord',),
    'any_of': ('anyof', 'any_of', 'anyOf', 'oneOf', 'one_of',),
    'any_except': ('anyExcept',),
    'any_char_except': ('anyCharExcept',),
    'printable_and_space': ('printableAndSpace',),
    'if_proceded_by': ('ifFollowedBy', 'if_followed_by', 'ifProcededBy',),
    'if_exists': ('ifExists',),
    'if_not_proceded_by': ('ifNotFollowedBy', 'if_not_followed_by', 'ifNotProcededBy'),
    'if_preceded_by': ('ifPrecededBy',),
    'if_not_preceded_by': ('ifNotPrecededBy',),
    'if_enclosed_with': ('ifEnclosedWith', 'if_enclosed_by', 'ifEnclosedBy',),
    'passive_group': ('passiveGroup',),
    'earlier_group': ('sameAs', 'same_as', 'earlierGroup', 'sameAsGroup', 'same_as_group',),
    'is_exactly': ('exactly', 'isExactly',),
    'optional': ('oneOrNone', 'one_or_none', 'opt',),
    'at_least_one': ('oneOrMore', 'one_or_more', 'atLeastOne', 'atLeast1', 'at_least_1',),
    'at_least_none': ('noneOrMore', 'none_or_more', 'atLeastNone', 'at_least_0', 'atLeast0', 'anyAmt', 'any_amt', 'zeroOrMore', 'zero_or_more',),
    'ASCII': ('ascii', 'a',),
    'DOTALL': ('dotall', 's',),
    'IGNORECASE': ('ignorecase', 'i', 'ignoreCase', 'ignore_case',),
    'LOCALE': ('locale', 'L',),
    'MULTILINE': ('multiline', 'm',),
    'signed': ('integer',),
    'literally_anything': ('literallyAnything',),
    'word_boundary': ('wordBoundary',),
    'not_word_boundary': ('notWordBoundary',),
    'rgroup': ('replaceGroup', 'replace_group',),
    'replace_entire': ('replaceAll', 'replace_all', 'replaceEntire',),
}



import re
import sys

from ezregex import EZRegex

# TODO: make all the flag functions here also accept re.FLAG types (internally they should work the same though)

class PythonEZRegex(EZRegex):
    _escape_chars=b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f'
    _repl_escape_chars=b''
    _compiled = None

    def compile(self, add_flags=True):
        return re.compile(self._compile(add_flags))

    @property
    def compiled(self):
        if self._compiled is None:
            self.__setattr__('_compiled', self.compile(), True)
        return self._compiled

    # Shadowing the re functions
    def search(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().search(string, pos, endpos)

    def match(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().match(string, pos, endpos)

    def fullmatch(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().fullmatch(string, pos, endpos)

    def split(self, string, maxsplit=0):
        return self.compile().split(string, maxsplit)

    def findall(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().findall(string, pos, endpos)

    def finditer(self, string, pos=0, endpos=sys.maxsize):
        return self.compile().finditer(string, pos, endpos)

    def sub(self, repl, string, count=0):
        return self.compile().sub(repl, string, count)

    def subn(self, repl, string, count=0):
        return self.compile().subn(repl, string, count)



#################### Maintained by Hatch ####################
# This file is auto-generated by hatch. If you'd like to customize this file
# please add your changes near the bottom marked for 'USER OVERRIDES'.
# EVERYTHING ELSE WILL BE OVERWRITTEN by hatch.
#############################################################
from io import open

from setuptools import find_packages, setup

with open('ezregex/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break
    else:
        version = '0.0.1'

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

REQUIRES = []

kwargs = {
    'name': 'ezregex',
    'version': version,
    'description': '',
    'long_description': readme,
    'author': 'Copeland Carter',
    'author_email': 'smartycope@gmail.com',
    'maintainer': 'Copeland Carter',
    'maintainer_email': 'smartycope@gmail.com',
    'url': 'https://github.com/smartycope/ezregex',
    'license': 'MIT/Apache-2.0',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    'install_requires': REQUIRES,
    'tests_require': ['coverage', 'pytest'],
    'packages': find_packages(exclude=('tests', 'tests.*')),

}

#################### BEGIN USER OVERRIDES ####################
# Add your customizations in this section.
###################### END USER OVERRIDES ####################

setup(
    **kwargs,
    long_description=readme,
    long_description_content_type='text/markdown',
    url='http://ezregex.org',
)
